/******************************************************************************
 * File: SpacesHandTrackingModule.cpp
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#include "SpacesHandTrackingExtension.h"
#include "Engine/Engine.h"
#include "IOpenXRARModule.h"
#include "SpacesCommon.h"
#include "OpenXRCore.h"
#include "IOpenXRARTrackedGeometryHolder.h"

#define LOCTEXT_NAMESPACE "FSpacesHandTrackingExtension"

static_assert(EHandKeypoint::Palm == static_cast<EHandKeypoint>(XR_HAND_JOINT_PALM_EXT), "EHandKeypoint enum does not match XrHandJointEXT.");
static_assert(EHandKeypoint::Wrist == static_cast<EHandKeypoint>(XR_HAND_JOINT_WRIST_EXT), "EHandKeypoint enum does not match XrHandJointEXT.");
static_assert(EHandKeypoint::ThumbMetacarpal == static_cast<EHandKeypoint>(XR_HAND_JOINT_THUMB_METACARPAL_EXT), "EHandKeypoint enum does not match XrHandJointEXT.");
static_assert(EHandKeypoint::IndexTip == static_cast<EHandKeypoint>(XR_HAND_JOINT_INDEX_TIP_EXT), "EHandKeypoint enum does not match XrHandJointEXT.");
static_assert(EHandKeypoint::LittleTip == static_cast<EHandKeypoint>(XR_HAND_JOINT_LITTLE_TIP_EXT), "EHandKeypoint enum does not match XrHandJointEXT.");

/** Constructor */
FSpacesHandTrackingExtension::FSpacesHandTrackingExtension() : HandDeviceIndex(0)
{
	IModularFeatures::Get().RegisterModularFeature(IMotionController::GetModularFeatureName(), static_cast<IMotionController*>(this));
	IModularFeatures::Get().RegisterModularFeature(IHandTracker::GetModularFeatureName(), static_cast<IHandTracker*>(this));
	
	UE_LOG(LogSpacesRuntime, Log, TEXT("Startup SpacesHandTrackingExtension"));
	RegisterOpenXRExtensionModularFeature();
}

/** Destructor */
FSpacesHandTrackingExtension::~FSpacesHandTrackingExtension()
{
	IModularFeatures::Get().UnregisterModularFeature(IMotionController::GetModularFeatureName(), static_cast<IMotionController*>(this));
	IModularFeatures::Get().UnregisterModularFeature(IHandTracker::GetModularFeatureName(), static_cast<IHandTracker*>(this));

	UE_LOG(LogSpacesRuntime, Log, TEXT("Shutdown SpacesHandTrackingExtension"));

	UnregisterOpenXRExtensionModularFeature();
}

/** ISpacesOpenXRExtension */
FString FSpacesHandTrackingExtension::GetDisplayName()
{
	return FString(TEXT("Spaces Hand Tracking"));
}

void FSpacesHandTrackingExtension::PostCreateInstance(XrInstance InInstance)
{
	SetInstanceHandle(InInstance);
}

void FSpacesHandTrackingExtension::PostCreateSession(XrSession InSession)
{
	SetSessionHandle(InSession);
}

bool FSpacesHandTrackingExtension::GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions)
{
	OutExtensions.Add(XR_EXT_HAND_TRACKING_EXTENSION_NAME);
	OutExtensions.Add(XR_QCOM_HAND_TRACKING_GESTURE_EXTENSION_NAME);
	OutExtensions.Add(XR_MSFT_HAND_TRACKING_MESH_EXTENSION_NAME);
	return true;
}

const void* FSpacesHandTrackingExtension::OnCreateSession(XrInstance InInstance, XrSystemId InSystem, const void* InNext)
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("OnCreateSession"));
	
	XrSystemHandTrackingMeshPropertiesMSFT HandMeshTrackingSystemProperties{ XR_TYPE_SYSTEM_HAND_TRACKING_MESH_PROPERTIES_MSFT };
	XrSystemHandTrackingPropertiesEXT HandTrackingSystemProperties{XR_TYPE_SYSTEM_HAND_TRACKING_PROPERTIES_EXT, &HandMeshTrackingSystemProperties};
	XrSystemProperties SystemPropertiesData{XR_TYPE_SYSTEM_PROPERTIES,&HandTrackingSystemProperties };
	
	XR_ENSURE(OpenXRDynamicAPI::xrGetSystemProperties(InInstance, InSystem, &SystemPropertiesData));
	
	bHandTrackingSupported = HandTrackingSystemProperties.supportsHandTracking == XR_TRUE;
	UE_LOG(LogSpacesRuntime, Log, TEXT("bHandTrackingSupported == %s"), bHandTrackingSupported ? TEXT("True") : TEXT("False"));
	
	if(bHandTrackingSupported)
	{
		for(FHandState& HandState : HandStates)
		{
			HandState.HandMeshIndicesMaxNum = HandMeshTrackingSystemProperties.maxHandMeshIndexCount;
			HandState.HandMeshVerticesMaxNum = HandMeshTrackingSystemProperties.maxHandMeshVertexCount;
			HandState.HandMeshGuid = FGuid::NewGuid();
		}
	}
	
	XRTrackingSystem = GEngine->XRSystem.Get();
	
	if (IOpenXRARModule::IsAvailable())
	{
		HandsTrackedMeshHolder = IOpenXRARModule::Get().GetTrackedMeshHolder();
		UE_LOG(LogSpacesRuntime, Log, TEXT("Obtained HandsTrackedMeshHolder (%p)"), HandsTrackedMeshHolder);
	}
	else
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("Cannot access to HandsTrackedMeshHolder"));
	}
	
	return InNext;
}

const void* FSpacesHandTrackingExtension::OnBeginSession(XrSession InSession, const void* InNext)
{
	SetSessionHandle(InSession);
	return InNext;
}

void FSpacesHandTrackingExtension::UpdateDeviceLocations(XrSession InSession, XrTime DisplayTime, XrSpace TrackingSpace)
{
	if (!bHandTrackingSupported || !bHandTrackingActive)
	{
		return;
	}

	for (int i = 0; i < 2; ++i)
	{
		UpdateHandTracking(i, DisplayTime, TrackingSpace);
		if(bHandMeshingActive)
		{
			UpdateHandMeshing(i, DisplayTime, TrackingSpace);
		}
		TryGetHandGestureData(i, DisplayTime);
	}
}

void FSpacesHandTrackingExtension::UpdateHandTracking(int HandIndex, XrTime DisplayTime, XrSpace TrackingSpace)
{
	FHandState& HandState = HandStates[HandIndex];

	XrHandJointsLocateInfoEXT HandJointsLocateInfo{ XR_TYPE_HAND_JOINTS_LOCATE_INFO_EXT };
	HandJointsLocateInfo.time = DisplayTime;
	HandJointsLocateInfo.baseSpace = TrackingSpace;
	
	if(SnapdragonSpaces::xrLocateHandJointsEXT(HandState.HandTracker, &HandJointsLocateInfo, &HandState.HandLocations) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("xrLocateHandJointsEXT failed"));
		return;
	}
	
	HandState.JointPosesAvailable = HandState.HandLocations.isActive == XR_TRUE;
	
	if (HandState.JointPosesAvailable) {
		const float WorldToMetersScalingFactor = XRTrackingSystem->GetWorldToMetersScale();
		
		static_assert(XR_HAND_JOINT_PALM_EXT == 0 && XR_HAND_JOINT_LITTLE_TIP_EXT == XR_HAND_JOINT_COUNT_EXT - 1, "This XrHandJointEXT enum is not supposed to be used here.");
		
		for (int i = 0; i < XR_HAND_JOINT_COUNT_EXT; ++i)
		{
			const XrHandJointLocationEXT& HandJointLocation = HandState.HandJointLocations[i];
			const FTransform JointTransform = ToFTransform(HandJointLocation.pose, WorldToMetersScalingFactor);
			HandState.HandKeypointTransforms[i] = JointTransform;
			HandState.Radii[i] = HandJointLocation.radius * WorldToMetersScalingFactor;
		}
	}
}

/** IMotionController interface */
bool FSpacesHandTrackingExtension::GetControllerOrientationAndPosition(const int32 ControllerIndex, const FName MotionSource, FRotator& OutOrientation, FVector& OutPosition, float WorldToMetersScale) const
{
	if (!bHandTrackingSupported || !bHandTrackingActive)
	{
		return false;
	}
	
	bool bIsHandControllerTracked = false;
	
	if (ControllerIndex == HandDeviceIndex)
	{
		const FHandState& HandState = (MotionSource == FName("EControllerHand::Left")) ? GetLeftHandState() : GetRightHandState();
		const FTransform HandControllerTransform = HandState.GetTransform(EHandKeypoint::Palm);
		bIsHandControllerTracked = HandState.JointPosesAvailable;
		OutPosition = HandControllerTransform.GetLocation();
		OutOrientation = HandControllerTransform.GetRotation().Rotator();
	}

	//Calling super for backward compatibility
	FXRMotionControllerBase::GetControllerOrientationAndPosition(ControllerIndex, MotionSource, OutOrientation, OutPosition, WorldToMetersScale);

	return bIsHandControllerTracked;
}

bool FSpacesHandTrackingExtension::GetControllerOrientationAndPosition(const int32 ControllerIndex, const EControllerHand DeviceHand, FRotator& OutOrientation, FVector& OutPosition, float WorldToMetersScale) const
{
	if (!bHandTrackingSupported || !bHandTrackingActive)
	{
		return false;
	}

	bool bIsHandControllerTracked = false;
	
	if (ControllerIndex == HandDeviceIndex && GetControllerTrackingStatus(ControllerIndex, DeviceHand) != ETrackingStatus::NotTracked)
	{
		const FHandState& HandState = (DeviceHand == EControllerHand::Left) ? GetLeftHandState() : GetRightHandState();
		const FTransform HandControllerTransform = HandState.GetTransform(EHandKeypoint::Palm);

		OutPosition = HandControllerTransform.GetLocation();
		OutOrientation = HandControllerTransform.GetRotation().Rotator();
		bIsHandControllerTracked = true;
	}

	return bIsHandControllerTracked;
}

ETrackingStatus FSpacesHandTrackingExtension::GetControllerTrackingStatus(const int32 ControllerIndex, const EControllerHand DeviceHand) const
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("FSpacesHandTrackingExtension: GetControllerTrackingStatus"));
	if (!bHandTrackingSupported)
	{
		return ETrackingStatus::NotTracked;
	}
	const FHandState& HandState = (DeviceHand == EControllerHand::Left) ? GetLeftHandState() : GetRightHandState();
	return HandState.JointPosesAvailable ? ETrackingStatus::Tracked : ETrackingStatus::NotTracked;
}

FName FSpacesHandTrackingExtension::GetMotionControllerDeviceTypeName() const
{
	const static FName DeviceTypeName(TEXT("OpenXRHandTracking"));
	return DeviceTypeName;
}

void FSpacesHandTrackingExtension::EnumerateSources(TArray<FMotionControllerSource>& SourcesOut) const
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("FSpacesHandTrackingExtension: EnumerateSources"));
	
	check(IsInGameThread());

	SourcesOut.Empty(EHandKeypointCount * 2);

	const UEnum* EHandKeypointPtr = FindObject<UEnum>(ANY_PACKAGE, TEXT("EHandKeypoint"), true);
	check(EHandKeypointPtr != nullptr);
	
	for (int32 i = 0; i < EHandKeypointCount; ++i)
	{
		const FString EnumString = EHandKeypointPtr->GetNameByValue(i).ToString();
		FString CleanEnumString = EnumString.Replace(TEXT("EHandKeypoint::"), TEXT(""));
		FString StringLeft = "Left" + CleanEnumString;
		FString StringRight = "Right" + CleanEnumString;
		SourcesOut.Add(FName(StringLeft));
		SourcesOut.Add(FName(StringRight));
	}
}

/** IHandTracker */
FName FSpacesHandTrackingExtension::GetHandTrackerDeviceTypeName() const
{
	return GetMotionControllerDeviceTypeName();
}

bool FSpacesHandTrackingExtension::IsHandTrackingStateValid() const
{
	return bHandTrackingSupported;
}

bool FSpacesHandTrackingExtension::GetKeypointState(EControllerHand Hand, EHandKeypoint Keypoint, FTransform& OutTransform, float& OutRadius) const
{
	if (!bHandTrackingSupported || !bHandTrackingActive)
	{
		return false;
	}

	bool HasTransform = false;

	const FHandState& HandState = (Hand == EControllerHand::Left) ? GetLeftHandState() : GetRightHandState();
	HasTransform = HandState.GetTransform(Keypoint, OutTransform);
	OutRadius = HandState.Radii[(uint32)Keypoint];

	if (HasTransform)
	{
		OutTransform *= XRTrackingSystem->GetTrackingToWorldTransform();
	}

	return HasTransform;
}

bool FSpacesHandTrackingExtension::GetAllKeypointStates(EControllerHand Hand, TArray<FVector>& OutPositions, TArray<FQuat>& OutRotations, TArray<float>& OutRadii) const
{
	if (!bHandTrackingSupported || !bHandTrackingActive)
	{
		return false;
	}

	const FHandState& HandState = (Hand == EControllerHand::Left) ? GetLeftHandState() : GetRightHandState();

	if (!HandState.JointPosesAvailable)
	{
		return false;
	}

	OutPositions.Empty(EHandKeypointCount);
	OutRotations.Empty(EHandKeypointCount);
	OutRadii.Empty(EHandKeypointCount);
	
	const FTransform TrackingToWorldTransform = XRTrackingSystem->GetTrackingToWorldTransform();

	for (int i = 0; i < EHandKeypointCount; ++i)
	{
		FTransform HandKeypointWorldTransform = HandState.HandKeypointTransforms[i] * TrackingToWorldTransform;
		OutPositions.Add(HandKeypointWorldTransform.GetLocation());
		OutRotations.Add(HandKeypointWorldTransform.GetRotation());
		OutRadii.Add(HandState.Radii[i]);
	}
	
	return true;
}

bool FSpacesHandTrackingExtension::HasHandMeshData() const
{
	if(!bHandMeshingActive)
	{
		return false;
	}
	
	for (int i = 0; i < 2; ++i)
	{
		const FHandState& HandState = HandStates[i];
		if (HandState.HandMeshVerticesNum > 0)
		{
			UE_LOG(LogSpacesRuntime, Log, TEXT("HasHandMeshData: True"));
			UE_LOG(LogSpacesRuntime, Log, TEXT("VerticesCount: %d"), HandState.HandMeshVerticesNum);
			return true;
		}
	}
	UE_LOG(LogSpacesRuntime, Log, TEXT("HasHandMeshData: False"));
	return false;
}

bool FSpacesHandTrackingExtension::GetHandMeshData(EControllerHand Hand, TArray<FVector>& OutVertices, TArray<FVector>& OutNormals, TArray<int32>& OutIndices, FTransform& OutHandMeshTransform) const
{
	check(IsInGameThread());
	
	if(!HasHandMeshData() || !bHandMeshingActive)
	{
		return false;
	}

	const FHandState& HandState = (Hand == EControllerHand::Left) ? HandStates[0] : HandStates[1];

	OutIndices.Reset(HandState.HandMeshIndicesNum * 2);
	OutVertices.Reset(HandState.HandMeshVerticesNum);
	OutNormals.Reset(HandState.HandMeshVerticesNum);

	OutIndices.AddUninitialized(HandState.HandMeshIndicesNum * 2);
	OutVertices.AddUninitialized(HandState.HandMeshVerticesNum);
	OutNormals.AddUninitialized(HandState.HandMeshVerticesNum);
	
	OutHandMeshTransform = HandState.HmLocalToTrackingTransform * XRTrackingSystem->GetTrackingToWorldTransform();
	
	auto Indices = OutIndices.GetData();
	auto OriginalIndices = HandState.HandMeshIndices.GetData();
	
	for(int i = 0; i < (HandState.HandMeshIndicesNum/3); ++i)
	{
		Indices[0] = OriginalIndices[2];
		Indices[1] = OriginalIndices[1];
		Indices[2] = OriginalIndices[0];

		Indices[3] = OriginalIndices[2];
		Indices[4] = OriginalIndices[1];
		Indices[5] = OriginalIndices[0];
		
		Indices += 6;
		OriginalIndices += 3;
	}
	
	for(int i = 0; i < HandState.HandMeshVerticesNum; ++i)
	{
		FVector& Vertex = OutVertices[i];
		FVector& Normal = OutNormals[i];
		const XrHandMeshVertexMSFT& OriginalVertex = HandState.HandMeshVertices[i];
		Vertex = ToFVector(OriginalVertex.position, XRTrackingSystem->GetWorldToMetersScale());
		Normal = ToFVector(OriginalVertex.normal);
	}
	
	return true;
}

/** SpacesHandTrackingModule */
const FSpacesHandTrackingExtension::FHandState& FSpacesHandTrackingExtension::GetLeftHandState() const
{
	return HandStates[0];
}

const FSpacesHandTrackingExtension::FHandState& FSpacesHandTrackingExtension::GetRightHandState() const
{
	return HandStates[1];
}

bool FSpacesHandTrackingExtension::GetHandTrackingSupported() const
{
	return bHandTrackingSupported;
}

void FSpacesHandTrackingExtension::TryGetHandGestureData(int HandIndex, XrTime DisplayTime)
{
	if (SnapdragonSpaces::xrGetHandGestureQCOM.IsValid() == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("xrGetHandGestureQCOM method not found!"));
		return;
	}
	
	FHandState& HandState = HandStates[HandIndex];

	if (HandState.HandTracker == XR_NULL_HANDLE)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("HandTracker deactivated."));
		return;
	}
	
	if(SnapdragonSpaces::xrGetHandGestureQCOM(HandState.HandTracker, DisplayTime, &HandState.HandGestureData) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed to get HandGesture data."));
	}
}

void FSpacesHandTrackingExtension::UpdateHandMeshing(int HandIndex, XrTime DisplayTime, XrSpace TrackingSpace)
{
	FHandState& HandState = HandStates[HandIndex];
	
	if(!HandState.JointPosesAvailable)
	{
		//If hand is not visible, remove meshes and return
		RemoveHandMesh(HandState);
		return;
	}
	
	if(LocateAndUpdateHandMesh(HandState, DisplayTime, TrackingSpace) == false || HandState.HandTrackingState != EARTrackingState::Tracking)
	{
		return;
	}
	FillHandMeshData(HandState);
}

bool FSpacesHandTrackingExtension::LocateAndUpdateHandMesh(FHandState& HandState, XrTime DisplayTime, XrSpace TrackingSpace)
{
	XrSpaceLocation HandMeshSpaceLocation { XR_TYPE_SPACE_LOCATION, nullptr, XrSpaceLocationFlags() };
	
	if(SnapdragonSpaces::xrLocateSpace(HandState.HandMeshSpace, TrackingSpace, DisplayTime, &HandMeshSpaceLocation) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed to xrLocateSpace."));
		return false;
	}
	
	if(SnapdragonSpaces::IsValidSpaceLocation(HandMeshSpaceLocation.locationFlags))
	{
		HandState.HandTrackingState = EARTrackingState::Tracking;
		XrHandMeshUpdateInfoMSFT HandMeshUpdateInfo { XR_TYPE_HAND_MESH_UPDATE_INFO_MSFT };
		HandMeshUpdateInfo.time = DisplayTime;
		HandMeshUpdateInfo.handPoseType = XR_HAND_POSE_TYPE_TRACKED_MSFT;

		XrHandMeshMSFT HandMesh { XR_TYPE_HAND_MESH_MSFT };
		HandMesh.indexBuffer.indexCapacityInput = HandState.HandMeshIndices.Num();
		HandMesh.indexBuffer.indices = HandState.HandMeshIndices.GetData();

		HandMesh.vertexBuffer.vertexCapacityInput = HandState.HandMeshVertices.Num();
		HandMesh.vertexBuffer.vertices = HandState.HandMeshVertices.GetData();

		if(SnapdragonSpaces::xrUpdateHandMeshMSFT(HandState.HandTracker, &HandMeshUpdateInfo, &HandMesh) == false)
		{
			UE_LOG(LogSpacesRuntime, Error, TEXT("Failed to xrUpdateHandMeshMSFT."));
			return false;
		}
		if (HandMesh.indexBufferChanged || HandMesh.vertexBufferChanged)
		{
			HandState.HandMeshIndicesNum = HandMesh.indexBuffer.indexCountOutput;
			HandState.HandMeshVerticesNum = HandMesh.vertexBuffer.vertexCountOutput;
		}
	}
	else
	{
		HandState.HandTrackingState = EARTrackingState::NotTracking;
		HandState.HandMeshIndicesNum = 0;
		HandState.HandMeshVerticesNum = 0;
	}
	
	HandState.HmLocalToTrackingTransform = ToFTransform(HandMeshSpaceLocation.pose, XRTrackingSystem->GetWorldToMetersScale());
	
	return true;
}

void FSpacesHandTrackingExtension::FillHandMeshData(FHandState& HandState)
{
	HandsTrackedMeshHolder->StartMeshUpdates();

	FOpenXRMeshUpdate* HandMeshUpdate = HandsTrackedMeshHolder->AllocateMeshUpdate(HandState.HandMeshGuid);
	HandMeshUpdate->Type = EARObjectClassification::HandMesh;
	HandMeshUpdate->TrackingState = HandState.HandTrackingState;
	HandMeshUpdate->LocalToTrackingTransform = HandState.HmLocalToTrackingTransform;
	
	HandMeshUpdate->Indices.AddUninitialized(HandState.HandMeshIndicesNum * 2);
	HandMeshUpdate->Vertices.AddUninitialized(HandState.HandMeshVerticesNum);
	
	auto Indices = HandMeshUpdate->Indices.GetData();
	auto OriginalIndices = HandState.HandMeshIndices.GetData();
	for(int i = 0; i < (HandState.HandMeshIndicesNum/3); ++i)
	{
		Indices[0] = OriginalIndices[2];
		Indices[1] = OriginalIndices[1];
		Indices[2] = OriginalIndices[0];

		Indices[3] = OriginalIndices[2];
		Indices[4] = OriginalIndices[1];
		Indices[5] = OriginalIndices[0];
		
		Indices += 6;
		OriginalIndices += 3;
	}
	
	for(int i = 0; i < HandState.HandMeshVerticesNum; ++i)
	{
		FVector& Vertex = HandMeshUpdate->Vertices[i];
		const XrHandMeshVertexMSFT& OriginalVertex = HandState.HandMeshVertices[i];
		Vertex = ToFVector(OriginalVertex.position, XRTrackingSystem->GetWorldToMetersScale());
	}
	HandsTrackedMeshHolder->EndMeshUpdates();
}

const XrHandGestureQCOM FSpacesHandTrackingExtension::GetHandGestureData(int HandIndex) const
{
	return HandStates[HandIndex].HandGestureData;
}

bool FSpacesHandTrackingExtension::HandStateJointPosesAvailable(int HandIndex) const
{
	return (HandStates[HandIndex].JointPosesAvailable && bHandTrackingActive);
}

void FSpacesHandTrackingExtension::SetHandTrackingState(bool active)
{
	bHandTrackingActive = active;
	int i = 0;
	for(FHandState& HandState : HandStates)
	{
		if(!active)
		{
			if(SnapdragonSpaces::xrDestroyHandTrackerEXT(HandState.HandTracker) == false)
			{
				UE_LOG(LogSpacesRuntime, Error, TEXT("Failed to destroy hand tracker: xrDestroyHandTrackerEXT failed."));
			}
			HandState.HandTracker = XR_NULL_HANDLE;
			RemoveHandMesh(HandState);
			UE_LOG(LogSpacesRuntime, Warning, TEXT("Hand tracker destroyed."));
		}
		else
		{
			if(HandState.HandTracker == XR_NULL_HANDLE)
			{
				InitHandState(GetSessionHandle(), i);
			}
		}
		i++;
	}
}

bool FSpacesHandTrackingExtension::GetSupportedCaptureTypes(TArray<EARCaptureType>& OutTypes)
{
	OutTypes.Add(EARCaptureType::HandMesh);
	return true;
}

bool FSpacesHandTrackingExtension::OnEnabledCaptureType(EARCaptureType InType)
{
	return true;
}

bool FSpacesHandTrackingExtension::OnDisabledCaptureType(EARCaptureType InType)
{
	return true;
}

bool FSpacesHandTrackingExtension::CanEnableExtension() const
{
	return bHandTrackingSupported;
}

bool FSpacesHandTrackingExtension::OnStart()
{
	UE_LOG(LogSpacesRuntime, Warning, TEXT("Hand Mesh Start."));
	bHandMeshingActive = true;
	return true;
}

void FSpacesHandTrackingExtension::OnPause()
{
	RemoveHandMeshes();
	bHandMeshingActive = false;
}

void FSpacesHandTrackingExtension::OnResume()
{
	bHandMeshingActive = true;
}

bool FSpacesHandTrackingExtension::OnStop()
{
	RemoveHandMeshes();
	bHandMeshingActive = false;
	return true;
}

void FSpacesHandTrackingExtension::SetHandMeshingStatus(bool Active)
{
	bHandMeshingActive = Active;
}

bool FSpacesHandTrackingExtension::GetHandMeshingStatus() const
{
	return bHandMeshingActive;
}

void FSpacesHandTrackingExtension::InitHandState(XrSession InSession, int HandIndex)
{
	if(!bHandTrackingSupported) return;
	if(XRTrackingSystem == nullptr) return;
	FHandState& HandState = HandStates[HandIndex];
	XrHandTrackerCreateInfoEXT HTCreateInfo{ XR_TYPE_HAND_TRACKER_CREATE_INFO_EXT };
	HTCreateInfo.handJointSet = XR_HAND_JOINT_SET_DEFAULT_EXT;
	HTCreateInfo.hand = XrHandEXT(XR_HAND_LEFT_EXT + HandIndex);
	if(SnapdragonSpaces::xrCreateHandTrackerEXT(InSession, &HTCreateInfo, &HandState.HandTracker) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed to create hand tracker: xrCreateHandTrackerEXT failed."));
	}

	XrHandMeshSpaceCreateInfoMSFT HandMeshSpaceCreateInfo{ XR_TYPE_HAND_MESH_SPACE_CREATE_INFO_MSFT };
	HandMeshSpaceCreateInfo.handPoseType = XR_HAND_POSE_TYPE_TRACKED_MSFT;
	HandMeshSpaceCreateInfo.poseInHandMeshSpace = ToXrPose(FTransform::Identity, XRTrackingSystem->GetWorldToMetersScale());
	if(SnapdragonSpaces::xrCreateHandMeshSpaceMSFT(HandState.HandTracker, &HandMeshSpaceCreateInfo, &HandState.HandMeshSpace) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed to create hand Mesh space: xrCreateHandMeshSpaceMSFT failed."));
	}
	
	HandState.HandMeshIndices.SetNum(HandState.HandMeshIndicesMaxNum);
	HandState.HandMeshVertices.SetNum(HandState.HandMeshVerticesMaxNum);
	UE_LOG(LogSpacesRuntime, Warning, TEXT("Hand tracker created."));
	bHandTrackingActive = true;
}

void FSpacesHandTrackingExtension::InitHandGestures()
{
	HandStates[0].HandGestureData = {XR_HAND_GESTURE_TYPE_UNKNOWN_QCOM,0.0,0.0}; 
	HandStates[1].HandGestureData = {XR_HAND_GESTURE_TYPE_UNKNOWN_QCOM,0.0,0.0}; 
}

void FSpacesHandTrackingExtension::RemoveHandMesh(const FHandState& HandState)
{
    if(HandsTrackedMeshHolder == nullptr)
    {
    	return;
    }
	
	HandsTrackedMeshHolder->RemoveMesh(HandState.HandMeshGuid);
}

void FSpacesHandTrackingExtension::RemoveHandMeshes()
{
	for(const FHandState& HandState : HandStates)
	{
		RemoveHandMesh(HandState);
	}
}

/** FHandState */
FSpacesHandTrackingExtension::FHandState::FHandState()
	: Radii{}
	, JointPosesAvailable(false)
	, HandMeshIndicesNum(0)
	, HandMeshVerticesNum(0)
	, HandMeshIndicesMaxNum(0)
	, HandMeshVerticesMaxNum(0)
{
	HandLocations.jointCount = XR_HAND_JOINT_COUNT_EXT;
	HandLocations.jointLocations = HandJointLocations;
	HandGestureData = {XR_HAND_GESTURE_TYPE_UNKNOWN_QCOM,0.0,0.0};
}

bool FSpacesHandTrackingExtension::FHandState::GetTransform(EHandKeypoint KeyPoint, FTransform& OutTransform) const
{
	int32 KeyPointNum = (int32)KeyPoint;
	check(KeyPointNum < EHandKeypointCount);
	OutTransform = HandKeypointTransforms[KeyPointNum];
	return JointPosesAvailable;
}

const FTransform& FSpacesHandTrackingExtension::FHandState::GetTransform(EHandKeypoint KeyPoint) const
{
	int32 KeyPointNum = (int32)KeyPoint;
	check(KeyPointNum < EHandKeypointCount);
	return HandKeypointTransforms[KeyPointNum];
}

#undef LOCTEXT_NAMESPACE
