﻿/******************************************************************************
 * File: SpacesMeshingExtension.cpp
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#include "SpacesMeshingExtension.h"


#include "IOpenXRARModule.h"
#include "SpacesCommon.h"
#include "IXRTrackingSystem.h"
#include "OpenXRCore.h"
#include "SpacesSessionConfig.h"
#include "Misc/FileHelper.h"
#include "PlaneDetection/SpacesPlaneDetectionExtension.h"
#include "Async/Async.h"
#include "SubSystems/SpacesEngineSubsystem.h"
#include "Kismet/KismetMathLibrary.h"

#define LOCTEXT_NAMESPACE "FSpacesMeshingExtension"

EARObjectClassification SnapdragonSpaces::ToARObjectClassification(XrSceneObjectTypeMSFT Type)
{
	switch (Type)
	{
	case XR_SCENE_OBJECT_TYPE_INFERRED_MSFT:
		return EARObjectClassification::World;
	case XR_SCENE_OBJECT_TYPE_UNCATEGORIZED_MSFT:
		return EARObjectClassification::World;
	case XR_SCENE_OBJECT_TYPE_BACKGROUND_MSFT:
		return EARObjectClassification::Unknown;
	case XR_SCENE_OBJECT_TYPE_WALL_MSFT:
		return EARObjectClassification::Wall;
	case XR_SCENE_OBJECT_TYPE_FLOOR_MSFT:
		return EARObjectClassification::Floor;
	case XR_SCENE_OBJECT_TYPE_CEILING_MSFT:
		return EARObjectClassification::Ceiling;
	case XR_SCENE_OBJECT_TYPE_PLATFORM_MSFT:
		return EARObjectClassification::Table;
	default:
		return EARObjectClassification::NotApplicable;
	}
}

EARPlaneOrientation SnapdragonSpaces::ToARPlaneOrientation(XrScenePlaneAlignmentTypeMSFT Type)
{
	switch (Type)
	{
	case XR_SCENE_PLANE_ALIGNMENT_TYPE_HORIZONTAL_MSFT:
		return EARPlaneOrientation::Horizontal;
	case XR_SCENE_PLANE_ALIGNMENT_TYPE_VERTICAL_MSFT:
		return EARPlaneOrientation::Vertical;
	case XR_SCENE_PLANE_ALIGNMENT_TYPE_NON_ORTHOGONAL_MSFT:
		return EARPlaneOrientation::Diagonal;
	default:
		return EARPlaneOrientation::Diagonal;
	}
}

FSpacesMeshingExtension::FSpacesMeshingExtension()
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Startup %s Extension"), *GetDisplayName());
	RegisterOpenXRExtensionModularFeature();
}

FSpacesMeshingExtension::~FSpacesMeshingExtension()
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Shutdown %s Extension started"), *GetDisplayName());
	UnregisterOpenXRExtensionModularFeature();
}

const void* FSpacesMeshingExtension::OnCreateSession(XrInstance InInstance, XrSystemId InSystem, const void* InNext)
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("OnCreateSession %s started"), *GetDisplayName());

	uint32 FeatureCount = 0;
	if(SnapdragonSpaces::xrEnumerateSceneComputeFeaturesMSFT(InInstance, InSystem, 0, &FeatureCount, nullptr) == true)
	{
		AvailableSceneComputeFeatures.SetNum(FeatureCount);
		if(SnapdragonSpaces::xrEnumerateSceneComputeFeaturesMSFT(InInstance, InSystem, FeatureCount, &FeatureCount, AvailableSceneComputeFeatures.GetData()) == false)
		{
			AvailableSceneComputeFeatures.Empty();
		}
		bDeviceSupportsMeshing = AvailableSceneComputeFeatures.Num() > 0;
		UE_LOG(LogSpacesRuntime, Verbose, TEXT("Got %d features"), AvailableSceneComputeFeatures.Num());
		for(const auto ComputeFeature : AvailableSceneComputeFeatures)
		{
			UE_LOG(LogSpacesRuntime, Verbose, TEXT("- %s"), SnapdragonSpaces::XrEnumStr(ComputeFeature));
		}
	}

	XRTrackingSystem = GEngine->XRSystem.Get();
	
	if (IOpenXRARModule::IsAvailable())
	{
		TrackedMeshHolder = IOpenXRARModule::Get().GetTrackedMeshHolder();
	}
	else
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Cannot access to TrackedMeshHolder"));
	}
	
	return InNext;
}

void FSpacesMeshingExtension::PostCreateSession(XrSession InSession)
{
	SetSessionHandle(InSession);
}

void FSpacesMeshingExtension::PostCreateInstance(XrInstance InInstance)
{
	SetInstanceHandle(InInstance);
}

FString FSpacesMeshingExtension::GetDisplayName()
{
	return FString(TEXT("Spaces Meshing"));
}

bool FSpacesMeshingExtension::GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions)
{
	OutExtensions.Add(XR_MSFT_SCENE_UNDERSTANDING_EXTENSION_NAME);
	return true;
}

void FSpacesMeshingExtension::UpdateDeviceLocations(XrSession InSession, XrTime DisplayTime, XrSpace TrackingSpace)
{
	if (!IsEnabled()) return;
    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Start UpdateDeviceLocations"));

	if(NormalsComputationFuture.IsReady())
	{
		const USpacesEngineSubsystem* SpacesEngineSubsystem = GEngine->GetEngineSubsystem<USpacesEngineSubsystem>();
		if(SpacesEngineSubsystem->OnSpacesSpatialMeshNormalsComputed.IsBound())
		{
			SpacesEngineSubsystem->OnSpacesSpatialMeshNormalsComputed.Broadcast(true);
		}
		NormalsComputationFuture.Reset();
		bLockVertices = false;
	}
	
	if(CurrentUpdateStep == ECurrentStep::ComputeScene){ ComputeScene(DisplayTime, TrackingSpace);}
	if(CurrentUpdateStep == ECurrentStep::ComputingScene){ ComputingScene();}
	if(CurrentUpdateStep == ECurrentStep::LoadScene){ LoadScene(DisplayTime, TrackingSpace);}
	if(CurrentUpdateStep == ECurrentStep::UpdateScene){ UpdateScene();}
}

bool FSpacesMeshingExtension::CanEnableExtension() const
{
	return bDeviceSupportsMeshing
	&& TrackedMeshHolder != nullptr
	;
}

bool FSpacesMeshingExtension::OnStart()
{
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("On Start %s Extension"), *GetDisplayName());
	return TryCreateSceneObserver();
}

void FSpacesMeshingExtension::OnPause()
{
}

void FSpacesMeshingExtension::OnResume()
{
}

bool FSpacesMeshingExtension::OnStop()
{
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("On Stop %s Extension"), *GetDisplayName());
	NormalsComputationFuture.Reset();
	CurrentUpdateStep = ECurrentStep::ComputeScene;
	SceneObserver.Clear();
	ClearAll();
	return true; 
}

TArray<FTrackedMeshHolder> FSpacesMeshingExtension::GetMeshes()
{
	UE_LOG(LogSpacesRuntime, Warning, TEXT("Num of meshes: %d"), MeshesArray.Num());
	return MeshesArray;
}

void FSpacesMeshingExtension::CalculateEnabledSceneComputeFeatures()
{
    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Start CalculateEnabledSceneComputeFeatures"));
	EnabledSceneComputeFeatures.Empty();
	const TArray<EARCaptureType>& CaptureTypes = GetEnabledCaptureTypes();
	if(CaptureTypes.Contains(EARCaptureType::SpatialMapping))
	{
		if(AvailableSceneComputeFeatures.Contains(XR_SCENE_COMPUTE_FEATURE_PLANE_MSFT))
		{
			EnabledSceneComputeFeatures.Add(XR_SCENE_COMPUTE_FEATURE_PLANE_MSFT);
			if(CurrentSessionConfig->bUseConvexHullDetection && AvailableSceneComputeFeatures.Contains(XR_SCENE_COMPUTE_FEATURE_PLANE_MESH_MSFT))
			{
				EnabledSceneComputeFeatures.Add(XR_SCENE_COMPUTE_FEATURE_PLANE_MESH_MSFT);
			}
		}
	}
	if(CaptureTypes.Contains(EARCaptureType::SceneUnderstanding))
	{
		if(AvailableSceneComputeFeatures.Contains(XR_SCENE_COMPUTE_FEATURE_VISUAL_MESH_MSFT))
		{
			EnabledSceneComputeFeatures.Add(XR_SCENE_COMPUTE_FEATURE_VISUAL_MESH_MSFT);
		}
		if(AvailableSceneComputeFeatures.Contains(XR_SCENE_COMPUTE_FEATURE_COLLIDER_MESH_MSFT))
		{
			EnabledSceneComputeFeatures.Add(XR_SCENE_COMPUTE_FEATURE_COLLIDER_MESH_MSFT);
		}
	}
}

bool FSpacesMeshingExtension::TryCreateSceneObserver()
{
    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Start TryCreateSceneObserver"));
	if(SceneObserver.IsValid()) return true;
	
	XrSceneObserverCreateInfoMSFT CreateInfo{
		XR_TYPE_SCENE_OBSERVER_CREATE_INFO_MSFT
	};

	SceneObserver = SnapdragonSpaces::FSceneObserverMSFTHandle(GetSessionHandle(), &CreateInfo);
	CalculateEnabledSceneComputeFeatures();
	
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("SceneObserver created is valid: %s"), SnapdragonSpaces::BoolToStr(SceneObserver.IsValid()));
	return SceneObserver.IsValid();
}

bool FSpacesMeshingExtension::TryReadSceneMeshBuffers(const XrSceneMSFT SceneHandle, const uint64_t MeshBufferId, const XrBool32 Supports16BitIndices, TArray<FVector>& Vertices, TArray<MRMESH_INDEX_TYPE>& Indices) const
{
    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Start TryReadSceneMeshBuffers"));
	constexpr bool bMRMeshUses16BitIndices = (sizeof(MRMESH_INDEX_TYPE) == sizeof(uint16));
	const bool bSupports16BitIndices = (Supports16BitIndices == XR_TRUE);
	
	if(!bSupports16BitIndices && bMRMeshUses16BitIndices)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("MRMesh needs 16 bit indices and current buffer only supports 32 bit indices!"));
		return false;
	}
	
	const float WorldToMetersScale = XRTrackingSystem->GetWorldToMetersScale();

	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Reading mesh buffers for %llu"), MeshBufferId);

	const XrSceneMeshBuffersGetInfoMSFT Info{
		XR_TYPE_SCENE_MESH_BUFFERS_GET_INFO_MSFT,
		nullptr,
		MeshBufferId
	};

	XrSceneMeshBuffersMSFT Buffers{
		XR_TYPE_SCENE_MESH_BUFFERS_MSFT
	};

	XrSceneMeshVertexBufferMSFT VertexBuffer{
		XR_TYPE_SCENE_MESH_VERTEX_BUFFER_MSFT
	};

	XrSceneMeshIndicesUint16MSFT Indices16Buffer{
		XR_TYPE_SCENE_MESH_INDICES_UINT16_MSFT
	};

	XrSceneMeshIndicesUint32MSFT Indices32Buffer{
		XR_TYPE_SCENE_MESH_INDICES_UINT32_MSFT
	};

	Buffers.next = &VertexBuffer;

	if constexpr (bMRMeshUses16BitIndices)
	{
		VertexBuffer.next = static_cast<void*>(&Indices16Buffer);
	}
	else
	{
		VertexBuffer.next = static_cast<void*>(&Indices32Buffer);
	}
	
	SnapdragonSpaces::xrGetSceneMeshBuffersMSFT(SceneHandle,&Info,&Buffers);

	const uint32_t VertexCount = VertexBuffer.vertexCountOutput;

	TArray<XrVector3f> NativeVertices;
	NativeVertices.SetNum(VertexCount);

	VertexBuffer.vertexCapacityInput = VertexCount;
	VertexBuffer.vertices = NativeVertices.GetData();

	uint32_t IndexCount;
	if constexpr (bMRMeshUses16BitIndices)
	{
		IndexCount = Indices16Buffer.indexCountOutput;
		Indices.SetNum(IndexCount);
		Indices16Buffer.indexCapacityInput = IndexCount;
		Indices16Buffer.indices = reinterpret_cast<uint16*>(Indices.GetData());
	}
	else
	{
		IndexCount = Indices32Buffer.indexCountOutput;
		Indices.SetNum(IndexCount);
		Indices32Buffer.indexCapacityInput = IndexCount;
		Indices32Buffer.indices = reinterpret_cast<uint32*>(Indices.GetData());
	}
	
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Got %u vertices"), VertexCount);
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Got %u indices"), IndexCount);
	
	SnapdragonSpaces::xrGetSceneMeshBuffersMSFT(SceneHandle,&Info,&Buffers);

	Vertices.Reserve(VertexCount);
	for(const auto& Vertex : NativeVertices)
	{
		Vertices.Add(ToFVector(Vertex, WorldToMetersScale));
	}
	
	return true;
}

template <XrSceneComponentTypeMSFT Value, typename Type>
bool FSpacesMeshingExtension::TryGetComponents(XrSceneMSFT Scene, TArray<XrSceneComponentMSFT>& SceneComponents,
                                            TArray<Type>& TypeComponents) const
{
    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Start TryGetComponents (%s)"), SnapdragonSpaces::XrEnumStr(Value));
	constexpr XrSceneComponentsGetInfoMSFT GetInfo{
		XR_TYPE_SCENE_COMPONENTS_GET_INFO_MSFT,
		nullptr,
		Value
	};

	XrSceneComponentsMSFT SceneComponentsMSFT{
		XR_TYPE_SCENE_COMPONENTS_MSFT,
		nullptr,0,0		
	};

	if(SnapdragonSpaces::xrGetSceneComponentsMSFT(Scene, &GetInfo, &SceneComponentsMSFT) == false)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Failed to get scene components count for %s"), SnapdragonSpaces::XrEnumStr(Value));
		return false;
	}
	const uint32_t Count = SceneComponentsMSFT.componentCountOutput;

	SceneComponents.SetNum(Count);
	SceneComponentsMSFT.componentCapacityInput = Count;
	SceneComponentsMSFT.components = SceneComponents.GetData();

	TypeComponents.SetNum(Count);
	if constexpr (Value == XR_SCENE_COMPONENT_TYPE_OBJECT_MSFT)
	{
		static_assert(TAreTypesEqual<Type, XrSceneObjectMSFT>::Value,"XR_SCENE_COMPONENT_TYPE_OBJECT_MSFT requires XrSceneObjectMSFT");
		XrSceneObjectsMSFT SceneObjects{
			XR_TYPE_SCENE_OBJECTS_MSFT,
			nullptr,
			Count,
			TypeComponents.GetData()
		};
		SceneComponentsMSFT.next = &SceneObjects;
		if(SnapdragonSpaces::xrGetSceneComponentsMSFT(Scene, &GetInfo, &SceneComponentsMSFT) == false)
		{
			UE_LOG(LogSpacesRuntime, Warning, TEXT("Failed to get scene components for %s"), SnapdragonSpaces::XrEnumStr(Value));
			return false;
		}
	}
	else if constexpr (Value == XR_SCENE_COMPONENT_TYPE_PLANE_MSFT)
	{
		static_assert(TAreTypesEqual<Type, XrScenePlaneMSFT>::Value,"XR_SCENE_COMPONENT_TYPE_PLANE_MSFT requires XrScenePlaneMSFT");
		XrScenePlanesMSFT ScenePlanes{
			XR_TYPE_SCENE_PLANES_MSFT,
			nullptr,
			Count,
			TypeComponents.GetData()
		};
		SceneComponentsMSFT.next = &ScenePlanes;
		if(SnapdragonSpaces::xrGetSceneComponentsMSFT(Scene, &GetInfo, &SceneComponentsMSFT) == false)
		{
			UE_LOG(LogSpacesRuntime, Warning, TEXT("Failed to get scene components for %s"), SnapdragonSpaces::XrEnumStr(Value));
			return false;
		}
	}
	else if constexpr (Value == XR_SCENE_COMPONENT_TYPE_VISUAL_MESH_MSFT || Value == XR_SCENE_COMPONENT_TYPE_COLLIDER_MESH_MSFT)
	{
		static_assert(TAreTypesEqual<Type, XrSceneMeshMSFT>::Value,"XR_SCENE_COMPONENT_TYPE_VISUAL_MESH_MSFT and XR_SCENE_COMPONENT_TYPE_COLLIDER_MESH_MSFT requires XrSceneMeshMSFT");
		XrSceneMeshesMSFT SceneMeshes{
			XR_TYPE_SCENE_MESHES_MSFT,
			nullptr,
			Count,
			TypeComponents.GetData()
		};
		SceneComponentsMSFT.next = &SceneMeshes;
		if(SnapdragonSpaces::xrGetSceneComponentsMSFT(Scene, &GetInfo, &SceneComponentsMSFT) == false)
		{
			UE_LOG(LogSpacesRuntime, Warning, TEXT("Failed to get scene components for %s"), SnapdragonSpaces::XrEnumStr(Value));
			return false;
		}
	}
	return true;
}

template bool FSpacesMeshingExtension::TryGetComponents<XR_SCENE_COMPONENT_TYPE_OBJECT_MSFT, XrSceneObjectMSFT>(
	XrSceneMSFT, TArray<XrSceneComponentMSFT>&, TArray<XrSceneObjectMSFT>&) const;
template bool FSpacesMeshingExtension::TryGetComponents<XR_SCENE_COMPONENT_TYPE_PLANE_MSFT, XrScenePlaneMSFT>(
	XrSceneMSFT, TArray<XrSceneComponentMSFT>&, TArray<XrScenePlaneMSFT>&) const;
template bool FSpacesMeshingExtension::TryGetComponents<XR_SCENE_COMPONENT_TYPE_VISUAL_MESH_MSFT, XrSceneMeshMSFT>(
	XrSceneMSFT, TArray<XrSceneComponentMSFT>&, TArray<XrSceneMeshMSFT>&) const;
template bool FSpacesMeshingExtension::TryGetComponents<XR_SCENE_COMPONENT_TYPE_COLLIDER_MESH_MSFT, XrSceneMeshMSFT>(
	XrSceneMSFT, TArray<XrSceneComponentMSFT>&, TArray<XrSceneMeshMSFT>&) const;

bool FSpacesMeshingExtension::TryGenerateObjectMap(const XrSceneMSFT Scene, TMap<FGuid, XrSceneObjectMSFT>& ObjectMap) const
{
    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Start TryGenerateObjectMap"));
	TArray<XrSceneComponentMSFT> Components;
	TArray<XrSceneObjectMSFT> Objects;
	if(!TryGetComponents<XR_SCENE_COMPONENT_TYPE_OBJECT_MSFT>(Scene, Components, Objects)){return false;}

	ObjectMap.Reserve(Objects.Num());
	for (int32 Index = 0; Index < Objects.Num(); Index++)
	{
		UE_LOG(LogSpacesRuntime, Verbose, TEXT("Found object with id: %s"), *SnapdragonSpaces::As<FGuid>(Components[Index].id).ToString());
		UE_LOG(LogSpacesRuntime, Verbose, TEXT("Object classification: %s"), SnapdragonSpaces::XrEnumStr(Objects[Index].objectType));
		ObjectMap.Add(SnapdragonSpaces::As<FGuid>(Components[Index].id), Objects[Index]);
	}
	return true;
}

bool FSpacesMeshingExtension::TryGeneratePlanesMap(const XrSceneMSFT Scene, const TMap<FGuid, XrSceneObjectMSFT>& ObjectMap)
{
    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Start TryGeneratePlanesMap"));
	const float WorldToMetersScale = XRTrackingSystem->GetWorldToMetersScale();
	
	TArray<XrSceneComponentMSFT> Components;
	TArray<XrScenePlaneMSFT> Planes;
	
	if(!TryGetComponents<XR_SCENE_COMPONENT_TYPE_PLANE_MSFT>(Scene, Components, Planes)){return false;}
	
	PlanesMap.Empty();
	PlanesMap.Reserve(Planes.Num());
	for (int32_t Index = 0; Index < Planes.Num(); Index++)
	{
		FTrackedPlaneHolder TrackedPlane;

		TrackedPlane.Id = SnapdragonSpaces::As<FGuid>(Components[Index].id);
		TrackedPlane.ParentId = SnapdragonSpaces::As<FGuid>(Components[Index].parentId);
		TrackedPlane.Classification = EARObjectClassification::NotApplicable;
		TrackedPlane.SpatialMeshUsageFlags = static_cast<EARSpatialMeshUsageFlags>(static_cast<__underlying_type(EARSpatialMeshUsageFlags)>(EARSpatialMeshUsageFlags::Visible) | static_cast<__underlying_type(EARSpatialMeshUsageFlags)>(EARSpatialMeshUsageFlags::Collision));
		TrackedPlane.Orientation = SnapdragonSpaces::ToARPlaneOrientation(Planes[Index].alignment);
		TrackedPlane.Extents = FVector(ToFVector2D(Planes[Index].size, WorldToMetersScale),0)*0.5f;
		TrackedPlane.TrackingState = EARTrackingState::Tracking;
		
		UE_LOG(LogSpacesRuntime, Verbose, TEXT("Found plane with id: %s"), *TrackedPlane.Id.ToString());
		UE_LOG(LogSpacesRuntime, Verbose, TEXT("Plane parent id: %s"), *TrackedPlane.ParentId.ToString());

		//TODO(AS): Remove true when 3DR layer fixes planes extents
	    if(CurrentSessionConfig->bUseConvexHullDetection || true)
	    {
	    	TryReadSceneMeshBuffers(Scene,Planes[Index].meshBufferId, Planes[Index].supportsIndicesUint16, TrackedPlane.Vertices, TrackedPlane.Indices);
	    	TrackedPlane.Extents = FBox(TrackedPlane.Vertices).GetExtent();
	    }
		
		PlanesMap.Add(TrackedPlane.Id, MoveTemp(TrackedPlane));
		
	}

	return true;
}

bool FSpacesMeshingExtension::TryGenerateMeshesMap(const XrSceneMSFT Scene, const TMap<FGuid, XrSceneObjectMSFT>& ObjectMap)
{
    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Start TryGenerateMeshesMap"));
	TArray<XrSceneComponentMSFT> Components;
	TArray<XrSceneMeshMSFT> Meshes;

	if(!TryGetComponents<XR_SCENE_COMPONENT_TYPE_VISUAL_MESH_MSFT>(Scene, Components, Meshes)){return false;}
	if(!bLockVertices && bComputeNormals){ MeshesArray.Empty(); }
		
	MeshesMap.Empty();
	MeshesMap.Reserve(Meshes.Num());
	for (int32_t Index = 0; Index < Meshes.Num(); Index++)
	{
		FTrackedMeshHolder TrackedMesh;

		TrackedMesh.Id = SnapdragonSpaces::As<FGuid>(Components[Index].id);
		TrackedMesh.ParentId = SnapdragonSpaces::As<FGuid>(Components[Index].parentId);
		if(const auto* ObjectMesh = ObjectMap.Find(TrackedMesh.Id); ObjectMesh != nullptr){
			TrackedMesh.Classification = SnapdragonSpaces::ToARObjectClassification(ObjectMesh->objectType);
		}
		else{
			TrackedMesh.Classification = EARObjectClassification::Unknown;
		}
		TrackedMesh.SpatialMeshUsageFlags = EARSpatialMeshUsageFlags::Visible;

		UE_LOG(LogSpacesRuntime, Verbose, TEXT("Found mesh with id: %s"), *TrackedMesh.Id.ToString());
		UE_LOG(LogSpacesRuntime, Verbose, TEXT("Mesh parent id: %s"), *TrackedMesh.ParentId.ToString());
		if(TryReadSceneMeshBuffers(Scene,Meshes[Index].meshBufferId, Meshes[Index].supportsIndicesUint16, TrackedMesh.Vertices, TrackedMesh.Indices))
		{
			if(!bLockVertices && bComputeNormals)
			{
				MeshesArray.Add(TrackedMesh);
				UE_LOG(LogSpacesRuntime, Verbose, TEXT("Number of vertices: %d"), TrackedMesh.Vertices.Num());
				UE_LOG(LogSpacesRuntime, Verbose, TEXT("Number of indices: %d"), TrackedMesh.Indices.Num());
			}
			MeshesMap.Add(TrackedMesh.Id, MoveTemp(TrackedMesh));
		}
	}
	
	if(!bLockVertices && bComputeNormals){ AsyncComputeVerticesNormals(); }
	
	return true;
}

bool FSpacesMeshingExtension::TryGenerateCollidersMap(const XrSceneMSFT Scene,const TMap<FGuid, XrSceneObjectMSFT>& ObjectMap)
{
    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Start TryGenerateCollidersMap"));
	TArray<XrSceneComponentMSFT> Components;
	TArray<XrSceneMeshMSFT> Meshes;

	if(!TryGetComponents<XR_SCENE_COMPONENT_TYPE_COLLIDER_MESH_MSFT>(Scene, Components, Meshes)){return false;}
	
	CollidersMap.Empty();
	CollidersMap.Reserve(Meshes.Num());
	for (int32_t Index = 0; Index < Meshes.Num(); Index++)
	{
		FTrackedMeshHolder TrackedMesh;

		TrackedMesh.Id = SnapdragonSpaces::As<FGuid>(Components[Index].id);
		TrackedMesh.ParentId = SnapdragonSpaces::As<FGuid>(Components[Index].parentId);
		TrackedMesh.Classification = ObjectMap.Find(TrackedMesh.Id) ? SnapdragonSpaces::ToARObjectClassification(ObjectMap[TrackedMesh.Id].objectType) : EARObjectClassification::Unknown;
		TrackedMesh.SpatialMeshUsageFlags = EARSpatialMeshUsageFlags::Collision;
		
		UE_LOG(LogSpacesRuntime, Verbose, TEXT("Found collider with id: %s"), *TrackedMesh.Id.ToString());
		UE_LOG(LogSpacesRuntime, Verbose, TEXT("Collider parent id: %s"), *TrackedMesh.ParentId.ToString());
		if(TryReadSceneMeshBuffers(Scene,Meshes[Index].meshBufferId, Meshes[Index].supportsIndicesUint16, TrackedMesh.Vertices, TrackedMesh.Indices))
		{
			CollidersMap.Add(TrackedMesh.Id, MoveTemp(TrackedMesh));
		}
	}

	return true;
	
}

bool FSpacesMeshingExtension::TryLocateUuids(const XrTime DisplayTime, const XrSpace TrackingSpace, const XrSceneMSFT Scene)
{
    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Start TryLocateUuids"));
	TArray<FGuid> Uuids;
	TArray<FGuid> TempUuids;
	TArray<XrSceneComponentLocationMSFT> Locations;
	
	PlanesMap.GetKeys(TempUuids);
	Uuids+=TempUuids;
	MeshesMap.GetKeys(TempUuids);
	Uuids+=TempUuids;
	CollidersMap.GetKeys(TempUuids);
	Uuids+=TempUuids;
	
	const uint32_t Count = Uuids.Num(); 
	Locations.SetNum(Count);


	const XrSceneComponentsLocateInfoMSFT LocateInfo{
		XR_TYPE_SCENE_COMPONENTS_LOCATE_INFO_MSFT,
		nullptr,
		TrackingSpace,
		DisplayTime,
		Count,
		SnapdragonSpaces::As<XrUuidMSFT>(Uuids.GetData())
	};

	XrSceneComponentLocationsMSFT ComponentLocations{
		XR_TYPE_SCENE_COMPONENT_LOCATIONS_MSFT,
		nullptr,
		Count,
		Locations.GetData()
	};

	if(SnapdragonSpaces::xrLocateSceneComponentsMSFT(Scene, &LocateInfo, &ComponentLocations) == false)
	{
		return false;
	}

	for (uint32_t Index = 0; Index < Count; Index++)
	{
		const auto Uuid = Uuids[Index];
		FTrackedMeshHolder* TrackedMesh;
		if((TrackedMesh = PlanesMap.Find(Uuid)) != nullptr){
		    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Meshing UUID found in planes map. (%s)"), *Uuid.ToString());
		}
		else if((TrackedMesh = MeshesMap.Find(Uuid)) != nullptr){
		    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Meshing UUID found in planes meshes. (%s)"), *Uuid.ToString());
		}
		else if((TrackedMesh = CollidersMap.Find(Uuid)) != nullptr){
		    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Meshing UUID found in planes colliders. (%s)"), *Uuid.ToString());
		}
		else
		{
		    UE_LOG(LogSpacesRuntime, Warning, TEXT("Meshing UUID not found in any map, ignoring it. (%s)"), *Uuid.ToString());
			continue;
		}
	    //TODO(AS): Implement state when we get proper flags from runtime 
		TrackedMesh->TrackingState = EARTrackingState::Tracking;//SnapdragonSpaces::IsValidSpaceLocation(Locations[Index].flags) ? EARTrackingState::Tracking : EARTrackingState::StoppedTracking;
		if(TrackedMesh->TrackingState == EARTrackingState::Tracking)
		{
			const float WorldToMetersScale = XRTrackingSystem->GetWorldToMetersScale();
			TrackedMesh->LocalToTrackingTransform = ToFTransform(Locations[Index].pose, WorldToMetersScale);
			UE_LOG(LogSpacesRuntime, Warning, TEXT("Plane transform: (%s)"), *TrackedMesh->LocalToTrackingTransform.ToString());
		}
		else
		{
			TrackedMesh->LocalToTrackingTransform = FTransform::Identity;
		}
	}

	return true;
}

void FSpacesMeshingExtension::RemovePlanes(const TArray<FGuid>& MissingPlanesId) const
{
    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Start RemovePlanes"));
	
	for(const auto MissingPlane : MissingPlanesId)
	{
		TrackedMeshHolder->RemovePlane(MissingPlane);
	}
	
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Cleaned up %d planes"), MissingPlanesId.Num());
	
}

void FSpacesMeshingExtension::RemoveMeshes(const TArray<FGuid>& MissingMeshesId) const
{
    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Start RemoveMeshes"));
	
	for(const auto MissingMesh : MissingMeshesId)
	{
		TrackedMeshHolder->RemoveMesh(MissingMesh);
	}
	
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Cleaned up %d meshes"), MissingMeshesId.Num());
	
}

void FSpacesMeshingExtension::ClearAll()
{
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Cleaning up old meshes..."));
	TrackedMeshHolder->StartMeshUpdates();
	
	RemovePlanes(PreviousPlanes);
	RemoveMeshes(PreviousMeshes);
	RemoveMeshes(PreviousColliders);
	
	TrackedMeshHolder->EndMeshUpdates();
	
	PreviousPlanes.Empty();
	PreviousMeshes.Empty();
	PreviousColliders.Empty();
}

void FSpacesMeshingExtension::ComputeScene(const XrTime DisplayTime, const XrSpace TrackingSpace)
{
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Start ComputeScene"));
	if(FDateTime::Now() < NextComputeSceneTime)
	{
		UE_LOG(LogSpacesRuntime, Verbose, TEXT("Too soon, waiting..."));
		CurrentUpdateStep = ECurrentStep::ComputeScene;
		return;
	}
	
	const float WorldToMetersScale = XRTrackingSystem->GetWorldToMetersScale();

	NextComputeSceneTime = FDateTime::Now() + FTimespan::FromMilliseconds(500);
	
	SceneSphereBounds.Empty();
	SceneBoxBounds.Empty();
	for(const auto Bound : static_cast<USpacesSessionConfig*>(CurrentSessionConfig)->Bounds)
	{
		if(Bound.SphereRadius > 0)
		{
			SceneSphereBounds.Add(XrSceneSphereBoundMSFT{
				ToXrVector(Bound.Origin, WorldToMetersScale),
				static_cast<float>(Bound.SphereRadius)
			});
		}
		else
		{
			SceneBoxBounds.Add(XrSceneOrientedBoxBoundMSFT{
				{{0,0,0,1}, ToXrVector(Bound.Origin, WorldToMetersScale)},
				ToXrVector(Bound.BoxExtent, WorldToMetersScale)
			});
		}
	}

	SceneComputeInfo = XrNewSceneComputeInfoMSFT{
		XR_TYPE_NEW_SCENE_COMPUTE_INFO_MSFT,
		nullptr,
		static_cast<uint32>(EnabledSceneComputeFeatures.Num()),
		EnabledSceneComputeFeatures.GetData(),
		XR_SCENE_COMPUTE_CONSISTENCY_SNAPSHOT_COMPLETE_MSFT,
		{
			TrackingSpace,
			DisplayTime,
			static_cast<uint32>(SceneSphereBounds.Num()),
			SceneSphereBounds.GetData(),
			static_cast<uint32>(SceneBoxBounds.Num()),
			SceneBoxBounds.GetData(),
			0,
			nullptr
		}
	};

	if(SnapdragonSpaces::xrComputeNewSceneMSFT(*SceneObserver, &SceneComputeInfo) == false)
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("Failed to start computing scene"));
		CurrentUpdateStep = ECurrentStep::ComputeScene;
		return;
	}
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Start computing scene"));
	CurrentUpdateStep = ECurrentStep::ComputingScene;
	
}

void FSpacesMeshingExtension::ComputingScene()
{
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("ComputingScene"));
	XrSceneComputeStateMSFT SceneComputeState;
	if(SnapdragonSpaces::xrGetSceneComputeStateMSFT(*SceneObserver, &SceneComputeState) == false)
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("Get scene compute state failed (%s)"), SnapdragonSpaces::XrEnumStr(SnapdragonSpaces::xrGetSceneComputeStateMSFT.GetLastResult()));
		CurrentUpdateStep = ECurrentStep::ComputeScene;
		return;
	}
	
	if (SceneComputeState == XR_SCENE_COMPUTE_STATE_UPDATING_MSFT)
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("Computing scene..."));
		CurrentUpdateStep = ECurrentStep::ComputingScene;
		return;
	}
	
	if (SceneComputeState == XR_SCENE_COMPUTE_STATE_COMPLETED_WITH_ERROR_MSFT ||
		SceneComputeState == XR_SCENE_COMPUTE_STATE_NONE_MSFT)
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("Computing scene failed (%s)"), SnapdragonSpaces::XrEnumStr(SceneComputeState));
		CurrentUpdateStep = ECurrentStep::ComputeScene;
		return;
	}

	CurrentUpdateStep = ECurrentStep::LoadScene;
	
}

void FSpacesMeshingExtension::LoadScene(const XrTime DisplayTime, const XrSpace TrackingSpace)
{
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("LoadScene"));
	const float WorldToMetersScale = XRTrackingSystem->GetWorldToMetersScale();
	constexpr XrSceneCreateInfoMSFT CreateInfo{XR_TYPE_SCENE_CREATE_INFO_MSFT};
	const SnapdragonSpaces::FSceneMSFTHandle Scene(*SceneObserver, &CreateInfo);

	if(!Scene.IsValid())
	{
		CurrentUpdateStep = ECurrentStep::ComputeScene;
		return;
	}

	TMap<FGuid, XrSceneObjectMSFT> ObjectMap;
	if(TryGenerateObjectMap(*Scene, ObjectMap) == false)
	{
	    UE_LOG(LogSpacesRuntime, Warning, TEXT("Failed to generate Object Map"));
	}

	if(GetEnabledCaptureTypes().Contains(EARCaptureType::SpatialMapping))
	{
		if(TryGeneratePlanesMap(*Scene, ObjectMap) == false)
		{
	        UE_LOG(LogSpacesRuntime, Warning, TEXT("Failed to generate Planes Map"));
		}
	}
	
	if(GetEnabledCaptureTypes().Contains(EARCaptureType::SceneUnderstanding))
	{
		if(TryGenerateMeshesMap(*Scene, ObjectMap) == false)
		{
	        UE_LOG(LogSpacesRuntime, Warning, TEXT("Failed to generate Meshes Map"));
		}

	    if(TryGenerateCollidersMap(*Scene, ObjectMap) == false)
	    {
	        UE_LOG(LogSpacesRuntime, Warning, TEXT("Failed to generate Colliders Map"));
	    }
	}
	
	if(TryLocateUuids(DisplayTime,TrackingSpace,*Scene) == false)
	{
		CurrentUpdateStep = ECurrentStep::ComputeScene;
		return;
	}


	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Loaded %d objects:"), ObjectMap.Num());
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Loaded %d planes"), PlanesMap.Num());
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Loaded %d meshes"), MeshesMap.Num());
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Loaded %d colliders"), CollidersMap.Num());

	CurrentUpdateStep = ECurrentStep::UpdateScene;
	
}

void FSpacesMeshingExtension::UpdateScene()
{
	const float WorldToMetersScale = XRTrackingSystem->GetWorldToMetersScale();
	TArray<FGuid> MissingPlanes;
	TArray<FGuid> MissingMeshes;
	TArray<FGuid> MissingColliders;
	
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Calculate Missing Meshes"));
	{
		
		for(auto PreviousPlane : PreviousPlanes)
		{
			if(!PlanesMap.Contains(PreviousPlane))
			{
				MissingPlanes.Add(PreviousPlane);
			}
		}
		PlanesMap.GetKeys(PreviousPlanes);

		for(auto PreviousMesh : PreviousMeshes)
		{
			if(!MeshesMap.Contains(PreviousMesh))
			{
				MissingMeshes.Add(PreviousMesh);
			}
		}
		MeshesMap.GetKeys(PreviousMeshes);

		for(auto PreviousCollider : PreviousColliders)
		{
			if(!CollidersMap.Contains(PreviousCollider))
			{
				MissingColliders.Add(PreviousCollider);
			}
		}
		CollidersMap.GetKeys(PreviousColliders);
	}
	TrackedMeshHolder->StartMeshUpdates();
    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Generate Meshes"));
	{
		UE_LOG(LogSpacesRuntime, Verbose, TEXT("Generate planes"));
		for(auto& Plane : PlanesMap)
		{
			FTrackedPlaneHolder& PlaneDataHolder = Plane.Value;
			PlaneDataHolder.TransformToXYPlane();
			if(!CurrentSessionConfig->bUseConvexHullDetection /*PlaneDataHolder.Vertices.Num() == 0*/)
			{
	
				UE_LOG(LogSpacesRuntime, Verbose,TEXT("Allocating extents plane"));
			
				FOpenXRPlaneUpdate* PlaneUpdate = TrackedMeshHolder->AllocatePlaneUpdate(Plane.Key);
	
				PlaneUpdate->Id = PlaneDataHolder.Id;
				PlaneUpdate->SpatialMeshUsageFlags = PlaneDataHolder.SpatialMeshUsageFlags;
				PlaneUpdate->Extent = PlaneDataHolder.Extents;
				PlaneUpdate->Type = PlaneDataHolder.Classification;
			
				if(PlaneDataHolder.TrackingState == EARTrackingState::Tracking)
				{
					PlaneUpdate->LocalToTrackingTransform = PlaneDataHolder.LocalToTrackingTransform;
				
					PlaneUpdate->TrackingState = EARTrackingState::Tracking;
					UE_LOG(LogSpacesRuntime, Verbose, TEXT("Initializing %u plane TRACKING. Extent: %s | Location: %s | Rotation: %s"), Plane.Key.D, *PlaneUpdate->Extent.ToString(), *PlaneUpdate->LocalToTrackingTransform.GetLocation().ToString(), *PlaneUpdate->LocalToTrackingTransform.GetRotation().ToString());
				}
				else
				{
					PlaneUpdate->TrackingState = EARTrackingState::NotTracking;
					PlaneUpdate->LocalToTrackingTransform = FTransform::Identity;
					UE_LOG(LogSpacesRuntime, Verbose, TEXT("Initializing %u plane NOT TRACKING. Extent: %s"), Plane.Key.D, *PlaneUpdate->Extent.ToString());
				}
			}
			else
			{
			
				UE_LOG(LogSpacesRuntime, Verbose,TEXT("Allocating mesh plane"));
			
				FOpenXRMeshUpdate* PlaneUpdate = TrackedMeshHolder->AllocateMeshUpdate(Plane.Key);
				PlaneUpdate->Id = PlaneDataHolder.Id;
				PlaneUpdate->SpatialMeshUsageFlags = PlaneDataHolder.SpatialMeshUsageFlags;
				PlaneUpdate->Type = PlaneDataHolder.Classification;
			
				PlaneUpdate->Vertices = MoveTemp(PlaneDataHolder.Vertices);
				PlaneUpdate->Indices = MoveTemp(PlaneDataHolder.Indices);
	
				UE_LOG(LogSpacesRuntime, Verbose, TEXT("Vertices: %d"), PlaneUpdate->Vertices.Num())
				UE_LOG(LogSpacesRuntime, Verbose, TEXT("Indices: %d"), PlaneUpdate->Indices.Num())
			
				PlaneUpdate->SpatialMeshUsageFlags = PlaneDataHolder.SpatialMeshUsageFlags;
			
				if(PlaneDataHolder.TrackingState == EARTrackingState::Tracking)
				{
					PlaneUpdate->LocalToTrackingTransform = PlaneDataHolder.LocalToTrackingTransform;
					PlaneUpdate->TrackingState = EARTrackingState::Tracking;
					UE_LOG(LogSpacesRuntime, Verbose, TEXT("Initializing %u plane TRACKING. Vertices: %d | Location: %s | Rotation: %s"), Plane.Key.D, PlaneUpdate->Vertices.Num(), *PlaneUpdate->LocalToTrackingTransform.GetLocation().ToString(), *PlaneUpdate->LocalToTrackingTransform.GetRotation().ToString());
				}
				else
				{
					PlaneUpdate->TrackingState = EARTrackingState::NotTracking;
					PlaneUpdate->LocalToTrackingTransform = FTransform::Identity;
					UE_LOG(LogSpacesRuntime, Verbose, TEXT("Initializing %u plane NOT TRACKING. Vertices: %d"), Plane.Key.D, PlaneUpdate->Vertices.Num());
				}
			}
		}
	
		UE_LOG(LogSpacesRuntime, Verbose, TEXT("Generate meshes"));
		for(auto& Mesh : MeshesMap)
		{
			FTrackedMeshHolder& MeshDataHolder = Mesh.Value;

			UE_LOG(LogSpacesRuntime, Verbose,TEXT("Allocating mesh"));
		
			FOpenXRMeshUpdate* MeshUpdate = TrackedMeshHolder->AllocateMeshUpdate(Mesh.Key);
			MeshUpdate->Id = MeshDataHolder.Id;
			MeshUpdate->SpatialMeshUsageFlags = MeshDataHolder.SpatialMeshUsageFlags;
			MeshUpdate->Type = MeshDataHolder.Classification;
		
			MeshUpdate->Vertices = MoveTemp(MeshDataHolder.Vertices);
			MeshUpdate->Indices = MoveTemp(MeshDataHolder.Indices);

			UE_LOG(LogSpacesRuntime, Verbose, TEXT("Vertices: %d"), MeshUpdate->Vertices.Num())
			UE_LOG(LogSpacesRuntime, Verbose, TEXT("Indices: %d"), MeshUpdate->Indices.Num())
		
			MeshUpdate->SpatialMeshUsageFlags = MeshDataHolder.SpatialMeshUsageFlags;
		
			if(MeshDataHolder.TrackingState == EARTrackingState::Tracking)
			{
				MeshUpdate->LocalToTrackingTransform = MeshDataHolder.LocalToTrackingTransform;
				MeshUpdate->TrackingState = EARTrackingState::Tracking;
				UE_LOG(LogSpacesRuntime, Verbose, TEXT("Initializing %u mesh TRACKING. Vertices: %d | Location: %s | Rotation: %s"), Mesh.Key.D, MeshUpdate->Vertices.Num(), *MeshUpdate->LocalToTrackingTransform.GetLocation().ToString(), *MeshUpdate->LocalToTrackingTransform.GetRotation().ToString());
			}
			else
			{
				MeshUpdate->TrackingState = EARTrackingState::NotTracking;
				MeshUpdate->LocalToTrackingTransform = FTransform::Identity;
				UE_LOG(LogSpacesRuntime, Verbose, TEXT("Initializing %u mesh NOT TRACKING. Vertices: %d"), Mesh.Key.D, MeshUpdate->Vertices.Num());
			}
		}
	
		UE_LOG(LogSpacesRuntime, Verbose, TEXT("Generate colliders"));
		for(auto& Collider : CollidersMap)
		{
			FTrackedMeshHolder& ColliderDataHolder = Collider.Value;
	
			UE_LOG(LogSpacesRuntime, Verbose,TEXT("Allocating collider"));
		
			FOpenXRMeshUpdate* ColliderUpdate = TrackedMeshHolder->AllocateMeshUpdate(Collider.Key);
			ColliderUpdate->Id = ColliderDataHolder.Id;
			ColliderUpdate->SpatialMeshUsageFlags = ColliderDataHolder.SpatialMeshUsageFlags;
			ColliderUpdate->Type = ColliderDataHolder.Classification;
		
			ColliderUpdate->Vertices = MoveTemp(ColliderDataHolder.Vertices);
			ColliderUpdate->Indices = MoveTemp(ColliderDataHolder.Indices);
			ColliderUpdate->SpatialMeshUsageFlags = ColliderDataHolder.SpatialMeshUsageFlags;
		
			if(ColliderDataHolder.TrackingState == EARTrackingState::Tracking)
			{
				ColliderUpdate->LocalToTrackingTransform = ColliderDataHolder.LocalToTrackingTransform;
				ColliderUpdate->TrackingState = EARTrackingState::Tracking;
				UE_LOG(LogSpacesRuntime, Verbose, TEXT("Initializing %u collider TRACKING. Vertices: %d | Location: %s | Rotation: %s"), Collider.Key.D, ColliderUpdate->Vertices.Num(), *ColliderUpdate->LocalToTrackingTransform.GetLocation().ToString(), *ColliderUpdate->LocalToTrackingTransform.GetRotation().ToString());
			}
			else
			{
				ColliderUpdate->TrackingState = EARTrackingState::NotTracking;
				ColliderUpdate->LocalToTrackingTransform = FTransform::Identity;
				UE_LOG(LogSpacesRuntime, Verbose, TEXT("Initializing %u collider NOT TRACKING. Vertices: %d"), Collider.Key.D, ColliderUpdate->Vertices.Num());
			}
		}
	}
	
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Remove Old Meshes"));
	{
		RemovePlanes(MissingPlanes);
		RemoveMeshes(MissingMeshes);
		RemoveMeshes(MissingColliders);
	}
	// Unreal is not setting the mesh bounds properly, it's always using 0,0,0 as origin and HALF_WORLD_MAX as extents
	// See Engine\Source\Runtime\MRMesh\Private\MRMeshComponent.cpp:693 and Engine\Source\Runtime\MRMesh\Private\MRMeshComponent.cpp:1055
	TrackedMeshHolder->EndMeshUpdates();
	CurrentUpdateStep = ECurrentStep::ComputeScene;
}

TArray<FVector> FSpacesMeshingExtension::ComputeVerticesNormals(TArray<FVector> Vertices, TArray<uint32> Indices, TArray<FVector>& Normals)
{
    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Start ComputeVerticesNormals"));
	Normals.Empty();
	Normals.SetNum(Vertices.Num());
	TArray<int> NormalsOrder;
	NormalsOrder.SetNumZeroed(Vertices.Num());
	
	for(int i = 0; i < Indices.Num(); i+=3)
	{
		if(!(Vertices.IsValidIndex(Indices[i]) && Vertices.IsValidIndex(Indices[i+1]) && Vertices.IsValidIndex(Indices[i+2])))
		{
			UE_LOG(LogSpacesRuntime, Log, TEXT("Index Not valid in Vertices: %d"),Indices[i]);
			continue;
		}
		//Edges of triangle
		FVector Edge1 = Vertices[Indices[i + 1]] - Vertices[Indices[i]];
		FVector Edge2 = Vertices[Indices[i + 2]] - Vertices[Indices[i]];
		
		//Normal of the triangle
		const FVector PlaneNormal = FVector::CrossProduct(Edge1,Edge2);
		
		if(!(Normals.IsValidIndex(Indices[i]) && Normals.IsValidIndex(Indices[i+1]) && Normals.IsValidIndex(Indices[i+2])))
		{
			UE_LOG(LogSpacesRuntime, Log, TEXT("Index Not valid in Normals: %d"), Indices[i]);
			continue;
		}
		
		NormalsOrder[Indices[i]]++;
		NormalsOrder[Indices[i + 1]]++;
		NormalsOrder[Indices[i + 2]]++;
		
		//Vertex normal is the sum of the triangles normals the vertex is part of
		Normals[Indices[i]] = (((Normals[Indices[i]]*(NormalsOrder[Indices[i]]-1))/NormalsOrder[Indices[i]]) + (PlaneNormal/NormalsOrder[Indices[i]]));
		Normals[Indices[i + 1]] = (((Normals[Indices[i+1]]*(NormalsOrder[Indices[i+1]]-1))/NormalsOrder[Indices[i+1]]) + (PlaneNormal/NormalsOrder[Indices[i+1]]));
		Normals[Indices[i + 2]] = (((Normals[Indices[i+2]]*(NormalsOrder[Indices[i+2]]-1))/NormalsOrder[Indices[i+2]]) + (PlaneNormal/NormalsOrder[Indices[i+2]]));
	}
	
	for(FVector& Normal : Normals)
	{
		//Normalize for lightning
		Normal = -Normal;
		Normal.Normalize();
	}
	
	return Normals;
}

void FSpacesMeshingExtension::AsyncComputeVerticesNormals()
{
    UE_LOG(LogSpacesRuntime, Verbose, TEXT("Start AsyncComputeVerticesNormals"));
	bLockVertices = true;
	TPromise<void> Promise;
	NormalsComputationFuture = Promise.GetFuture();
	AsyncTask(ENamedThreads::AnyThread, [this, &PrevIndicesNum = PreviousIndicesNum, Promise = MoveTemp(Promise)]()mutable
	{
		for(FTrackedMeshHolder& Mesh : MeshesArray)
		{
			if(Mesh.Indices.Num() == PrevIndicesNum) continue;
			PrevIndicesNum = Mesh.Indices.Num();
			FSpacesMeshingExtension::ComputeVerticesNormals(Mesh.Vertices, Mesh.Indices, Mesh.Normals);
		}
		Promise.SetValue();
	});
}

void FSpacesMeshingExtension::SetComputeNormalsState(bool Active)
{
	bComputeNormals = Active;
}

bool FSpacesMeshingExtension::GetComputeNormalsState() const
{
	return bComputeNormals;
}


bool FSpacesMeshingExtension::GetSupportedCaptureTypes(TArray<EARCaptureType>& OutTypes)
{
	OutTypes.Append({EARCaptureType::SpatialMapping, EARCaptureType::SceneUnderstanding});
	return true;
}

bool FSpacesMeshingExtension::OnEnabledCaptureType(EARCaptureType InType)
{
	switch (InType)
	{
	case EARCaptureType::SpatialMapping:
		if(AvailableSceneComputeFeatures.Contains(XR_SCENE_COMPUTE_FEATURE_PLANE_MSFT))
		{
			return true;
		}
		break;
	case EARCaptureType::SceneUnderstanding:
		if(AvailableSceneComputeFeatures.Contains(XR_SCENE_COMPUTE_FEATURE_VISUAL_MESH_MSFT))
		{
			return true;
		}
		break;
	default:
		break;
	}
	return false;
}

bool FSpacesMeshingExtension::OnDisabledCaptureType(EARCaptureType InType)
{
	return true;
}


#undef LOCTEXT_NAMESPACE
