/******************************************************************************
 * File: SpacesPlaneDetectionExtension.cpp
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#include "SpacesPlaneDetectionExtension.h"

#include "SpacesCommon.h"

#include "IOpenXRARModule.h"
#include "IOpenXRARTrackedGeometryHolder.h"

#include "ARBlueprintLibrary.h"
#include "OpenXRCore.h"
#include "IXRTrackingSystem.h"
#include "SpacesSessionConfig.h"

#define LOCTEXT_NAMESPACE "FSpacesPlaneDetectionExtension"


void SnapdragonSpaces::XrPlaneExtentToUnrealPlane(const XrPlaneExtentQCOM& PlaneExtent, FVector& CenterOffset, FVector& Extents, float Scale)
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Converting XrPlaneExtentQCOM: X{min:%f, max:%f},Y{min:%f, max:%f}"),
		   PlaneExtent.extentX.min, PlaneExtent.extentX.max, PlaneExtent.extentY.min, PlaneExtent.extentY.max);

	const XrVector3f XrExtents = {
		(PlaneExtent.extentX.max - PlaneExtent.extentX.min) / 2,
		(PlaneExtent.extentY.max - PlaneExtent.extentY.min) / 2,
		0
	};
	Extents = ToFVector(XrExtents, Scale).RotateAngleAxis(-90, FVector::RightVector);
	
	UE_LOG(LogSpacesRuntime, Log, TEXT("Converted extents: X: %f, Y:%f"), Extents.X, Extents.Y);
	
	const XrVector3f XrCenterOffset = {
		 (PlaneExtent.extentX.max + PlaneExtent.extentX.min) / 2,
		-(PlaneExtent.extentY.max + PlaneExtent.extentY.min) / 2,
		0
	};
	CenterOffset = ToFVector(XrCenterOffset, Scale).RotateAngleAxis(-90, FVector::RightVector); 

	UE_LOG(LogSpacesRuntime, Log, TEXT("Calculated center offset: X: %f, Y:%f"), CenterOffset.X, CenterOffset.Y);
}

FTransform SnapdragonSpaces::CalculatePlaneTransform(const XrPosef& Pose, const FVector& CenterOffset, float Scale)
{
	const FTransform PoseToTransform = ToFTransform(Pose, Scale);
	return FTransform(CenterOffset) * PoseToTransform;
}

XrPlaneFilterQCOM SnapdragonSpaces::PlaneDetectionModeToXrPlaneFilter(EARPlaneDetectionMode Mode)
{
	switch(Mode)
	{
	case EARPlaneDetectionMode::HorizontalPlaneDetection:
		return XrPlaneFilterQCOM::XR_PLANE_FILTER_HORIZONTAL_QCOM;
	case EARPlaneDetectionMode::VerticalPlaneDetection:
		return XrPlaneFilterQCOM::XR_PLANE_FILTER_VERTICAL_QCOM;
	default:
		return XrPlaneFilterQCOM::XR_PLANE_FILTER_ANY_QCOM;
	}
}

EARPlaneOrientation SnapdragonSpaces::XrPlaneTypeToPlaneOrientation(XrPlaneTypeQCOM Type)
{
	switch(Type)
	{
	case XrPlaneTypeQCOM::XR_PLANE_TYPE_HORIZONTAL_UPWARD_QCOM:
	case XrPlaneTypeQCOM::XR_PLANE_TYPE_HORIZONTAL_DOWNWARD_QCOM:
		return EARPlaneOrientation::Horizontal;
	case XrPlaneTypeQCOM::XR_PLANE_TYPE_VERTICAL_QCOM:
		return EARPlaneOrientation::Vertical;
	default:
		return EARPlaneOrientation::Diagonal;
	}
}

bool FSpacesPlaneDetectionExtension::TryCreatePlaneDetection(EARPlaneDetectionMode Mode, bool EnableConvexHull)
{
	if(!SnapdragonSpaces::xrCreatePlaneDetectionQCOM.IsValid())
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Create Plane Detection has a NULL handle"));
		return false;
	}

	UE_LOG(LogSpacesRuntime, Log, TEXT("Creating Plane Detection, convex hull enabled: %s"), SnapdragonSpaces::BoolToStr(EnableConvexHull));
	
	const XrPlaneFilterQCOM PlaneFilter = SnapdragonSpaces::PlaneDetectionModeToXrPlaneFilter(Mode);
	XrPlaneDetectionCreateInfoQCOM CreateInfo{
		XR_TYPE_PLANE_DETECTION_CREATE_INFO_QCOM,
		nullptr,
		PlaneFilter,
		(XrBool32)EnableConvexHull
	};

	PlaneDetection = SnapdragonSpaces::FPlaneDetectionQCOMHandle(GetSessionHandle(), &CreateInfo);
	
	if(!PlaneDetection.IsValid())
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed to create Plane Detection"));
		return false;
	}
	
	return true;
}

TArray<XrVector3f> FSpacesPlaneDetectionExtension::GetPlaneVertices(uint32_t ConvexHullId) const
{
	if(ConvexHullId == 0)
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("GetPlaneVertices: No convex hull ID"));
		return TArray<XrVector3f>();
	}

	const XrPlaneConvexHullBufferInfoQCOM ConvexHullInfo{
		XR_TYPE_PLANE_CONVEX_HULL_BUFFER_INFO_QCOM,
		nullptr,
		ConvexHullId
	};
			
	uint32_t RequiredVerticesCapacity;
	XrPlaneConvexHullVertexBufferQCOM ConvexHullVertexBufferSizeQuery{
		XR_TYPE_PLANE_CONVEX_HULL_VERTEX_BUFFER_QCOM,
		nullptr,
		0,
		&RequiredVerticesCapacity
	};

	bool Result = SnapdragonSpaces::xrGetPlaneConvexHullVertexBufferQCOM(PlaneDetection.Value(), &ConvexHullInfo, &ConvexHullVertexBufferSizeQuery);

	if(Result == false)
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("GetPlaneVertices: Can't get vertex capacity input"));
		return TArray<XrVector3f>();
	}

	UE_LOG(LogSpacesRuntime, Log, TEXT("GetPlaneVertices: Requires %u vertices"), RequiredVerticesCapacity);

	if(RequiredVerticesCapacity == 0)
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("GetPlaneVertices: No requires vertices"));
		return TArray<XrVector3f>();
	}
	
	uint32_t RetrievedVerticesCount;
	TArray<XrVector3f> Vertices;
	Vertices.SetNum(RequiredVerticesCapacity);
			
	XrPlaneConvexHullVertexBufferQCOM ConvexHullVertexBufferQuery{
		XR_TYPE_PLANE_CONVEX_HULL_VERTEX_BUFFER_QCOM,
		nullptr,
		(uint32_t)Vertices.Num(),
		&RetrievedVerticesCount,
		Vertices.GetData()
	};
	
	Result = SnapdragonSpaces::xrGetPlaneConvexHullVertexBufferQCOM(PlaneDetection.Value(), &ConvexHullInfo,
		&ConvexHullVertexBufferQuery);
			
	if(Result == false)
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("GetPlaneVertices: Can't get plane vertices"));
		return TArray<XrVector3f>();
	}

	Vertices.SetNum(RetrievedVerticesCount,true);
	
	UE_LOG(LogSpacesRuntime, Log, TEXT("GetPlaneVertices: Obtained %u vertices"), RetrievedVerticesCount);
	UE_LOG(LogSpacesRuntime, Log, TEXT("GetPlaneVertices: Vertices size is %d"), Vertices.Num());
	
	return Vertices;
}

bool FSpacesPlaneDetectionExtension::LocatePlanes(XrSpace Space, XrTime Time, TMap<FGuid, FTrackedPlaneHolder>
                                               & LocatedPlanes) const
{
	LocatedPlanes.Reset();
	
	bool Result;
	
	XrPlaneDetectionStateQCOM PlaneDetectionState;
	Result = SnapdragonSpaces::xrGetPlaneDetectionStateQCOM(PlaneDetection.Value(), &PlaneDetectionState);
	if(Result == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("GetPlaneDetectionState: Can't get the plane detection state"));
		return false;
	}
	
	if(PlaneDetectionState != XR_PLANE_DETECTION_STATE_TRACKING_QCOM)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("The plane detection state is %s"), SnapdragonSpaces::XrEnumStr(PlaneDetectionState));
		return false;
	}


	const XrPlanesLocateInfoQCOM LocateInfo{
		XR_TYPE_PLANES_LOCATE_INFO_QCOM,
		nullptr,
		Space,
		Time
	};
	
	uint32_t NumPlanes;
	XrPlaneLocationsQCOM LocationSizeQuery{
		XR_TYPE_PLANE_LOCATIONS_QCOM,
		nullptr,
		0,
		&NumPlanes,
		nullptr
	};

	Result = SnapdragonSpaces::xrLocatePlanesQCOM(PlaneDetection.Value(), &LocateInfo, &LocationSizeQuery);
	if(Result == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("LocatePlanes: Can't locate planes at the first time"));
		return false;
	}

	if(NumPlanes == 0)
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("LocatePlanes: no planes detected"));
		return false;
	}

	
	TArray<XrPlaneLocationQCOM> LocationBuffer;
	LocationBuffer.SetNum(NumPlanes);
	
	
	XrPlaneLocationsQCOM LocationsPlaneQuery{
		XR_TYPE_PLANE_LOCATIONS_QCOM,
		nullptr,
		(uint32_t)LocationBuffer.Max(),
		&NumPlanes,
		LocationBuffer.GetData()
	};
	
	Result = SnapdragonSpaces::xrLocatePlanesQCOM(PlaneDetection.Value(), &LocateInfo, &LocationsPlaneQuery);

	if(Result == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("LocatePlanes: Can't locate planes at the second time"));
		return false;
	}

	if(*LocationsPlaneQuery.planeCountOutput == 0)
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("LocatePlanes: no planes detected"));
		return false;
	}
	
	const float WorldToMetersScale = XRTrackingSystem->GetWorldToMetersScale();
	
	for(uint32_t i = 0; i < *(LocationsPlaneQuery.planeCountOutput); ++i)
	{
		const XrPlaneLocationQCOM& PlaneLocation = LocationsPlaneQuery.planeLocations[i];
		FTrackedPlaneHolder PlaneHolder;
		FVector CenterOffset;

		PlaneHolder.Id = FGuid(0,0,0, PlaneLocation.id);
		SnapdragonSpaces::XrPlaneExtentToUnrealPlane(PlaneLocation.size, CenterOffset, PlaneHolder.Extents, WorldToMetersScale);
		PlaneHolder.Orientation = SnapdragonSpaces::XrPlaneTypeToPlaneOrientation(PlaneLocation.planeType);
		PlaneHolder.TrackingState = SnapdragonSpaces::ToARTrackingState(PlaneLocation.locationFlags);
		
		const uint64_t ConvexHullBufferId = PlaneLocation.convexHullBufferId;
		
		TArray<XrVector3f> Vertices = GetPlaneVertices(ConvexHullBufferId);
		TArray<MRMESH_INDEX_TYPE> Indices;
		TArray<FVector> TransformedVertices;
		if(Vertices.Num()>2)
		{
			Indices.Reserve((Vertices.Num()-2)*3);
			for(int vertexIndex=0; vertexIndex<Vertices.Num(); ++vertexIndex){
				const XrVector3f& Vertex = Vertices[vertexIndex];
				TransformedVertices.Add(ToFVector(Vertex, WorldToMetersScale).RotateAngleAxis(-90, FVector::RightVector));
				if(vertexIndex>=2)
				{
					Indices.Add(0);
					Indices.Add(vertexIndex - 1 );
					Indices.Add(vertexIndex);
				}
			}
		}

		if(TransformedVertices.Num() > 0)
		{
			PlaneHolder.LocalToTrackingTransform = ToFTransform(PlaneLocation.pose, WorldToMetersScale);
		}
		else
		{
			PlaneHolder.LocalToTrackingTransform = SnapdragonSpaces::CalculatePlaneTransform(PlaneLocation.pose, CenterOffset, WorldToMetersScale);
		}
		PlaneHolder.Vertices = MoveTemp(TransformedVertices);
		PlaneHolder.Indices = MoveTemp(Indices);
		LocatedPlanes.Add(PlaneHolder.Id, MoveTemp(PlaneHolder));		
	}
	
	return true;
}

void FSpacesPlaneDetectionExtension::CalculateMissingPlanes(const TMap<FGuid, FTrackedPlaneHolder>& LocatedPlanes,
                                                          TArray<FGuid>& MissingPlanesId)
{
	for(auto PreviousPlaneId : PreviousPlanesId)
	{
		if(!LocatedPlanes.Contains(PreviousPlaneId))
		{
			MissingPlanesId.Add(PreviousPlaneId);
		}
	}
	LocatedPlanes.GetKeys(PreviousPlanesId);
}

void FSpacesPlaneDetectionExtension::RemovePlanes(const TArray<FGuid>& MissingPlanesId) const
{
	if(TrackedMeshHolder == nullptr) return;
	
    TrackedMeshHolder->StartMeshUpdates();
	//UE_LOG(LogSpacesRuntime, Log, TEXT("Cleaning all previous geometries"));	
	for(const auto MissingPlaneId : MissingPlanesId)
	{
		TrackedMeshHolder->RemovePlane(MissingPlaneId);
	}
    TrackedMeshHolder->EndMeshUpdates();
}

void FSpacesPlaneDetectionExtension::AllocatePlanes(TMap<FGuid, FTrackedPlaneHolder>& Planes) const
{
	if(TrackedMeshHolder == nullptr) return;
	
	const float WorldToMetersScale = XRTrackingSystem->GetWorldToMetersScale();

	TMap<FGuid, FTrackedPlaneHolder> NewPlanes = MoveTemp(Planes);
	
    TrackedMeshHolder->StartMeshUpdates();

	for(auto& Plane : NewPlanes)
	{
		if(Plane.Key.IsValid())
		{
			if(FTrackedPlaneHolder& PlaneDataHolder = Plane.Value; PlaneDataHolder.Vertices.Num() == 0)
			{

				UE_LOG(LogSpacesRuntime, Log,TEXT("Allocating extents plane"));
				
				FOpenXRPlaneUpdate* PlaneUpdate = TrackedMeshHolder->AllocatePlaneUpdate(Plane.Key);

				PlaneUpdate->Id = PlaneDataHolder.Id;
				PlaneUpdate->Extent = PlaneDataHolder.Extents;
				PlaneUpdate->SpatialMeshUsageFlags = static_cast<EARSpatialMeshUsageFlags>(static_cast<__underlying_type(EARSpatialMeshUsageFlags)>(EARSpatialMeshUsageFlags::Visible) | static_cast<__underlying_type(EARSpatialMeshUsageFlags)>(EARSpatialMeshUsageFlags::Collision));
				
				if(PlaneDataHolder.TrackingState == EARTrackingState::Tracking)
				{
					PlaneUpdate->LocalToTrackingTransform = PlaneDataHolder.LocalToTrackingTransform;
					PlaneUpdate->TrackingState = EARTrackingState::Tracking;
					UE_LOG(LogSpacesRuntime, Log, TEXT("Initializing %u plane TRACKING. Extent: %s | Location: %s | Rotation: %s"), Plane.Key.D, *PlaneUpdate->Extent.ToString(), *PlaneUpdate->LocalToTrackingTransform.GetLocation().ToString(), *PlaneUpdate->LocalToTrackingTransform.GetRotation().ToString());
				}
				else
				{
					PlaneUpdate->TrackingState = EARTrackingState::NotTracking;
					PlaneUpdate->LocalToTrackingTransform = FTransform::Identity;
					UE_LOG(LogSpacesRuntime, Log, TEXT("Initializing %u plane NOT TRACKING. Extent: %s"), Plane.Key.D, *PlaneUpdate->Extent.ToString());
				}
			}
			else
			{
				
				UE_LOG(LogSpacesRuntime, Log,TEXT("Allocating convex hull plane"));
				
				FOpenXRMeshUpdate* PlaneUpdate = TrackedMeshHolder->AllocateMeshUpdate(Plane.Key);
				PlaneUpdate->Id = PlaneDataHolder.Id;
				PlaneUpdate->SpatialMeshUsageFlags = static_cast<EARSpatialMeshUsageFlags>(static_cast<__underlying_type(EARSpatialMeshUsageFlags)>(EARSpatialMeshUsageFlags::Visible) | static_cast<__underlying_type(EARSpatialMeshUsageFlags)>(EARSpatialMeshUsageFlags::Collision));
				
				PlaneUpdate->Vertices = MoveTemp(PlaneDataHolder.Vertices);
				PlaneUpdate->Indices = MoveTemp(PlaneDataHolder.Indices);

				UE_LOG(LogSpacesRuntime, Log, TEXT("Vertices: %d"), PlaneUpdate->Vertices.Num())
				UE_LOG(LogSpacesRuntime, Log, TEXT("Indices: %d"), PlaneUpdate->Indices.Num())				
				
				if(PlaneDataHolder.TrackingState == EARTrackingState::Tracking)
				{
					PlaneUpdate->LocalToTrackingTransform = PlaneDataHolder.LocalToTrackingTransform;
					//PlaneUpdate->LocalToTrackingTransform = ToFTransform(PlaneData.pose, WorldToMetersScale);//SnapdragonSpaces::CalculatePlaneTransform(PlaneData.pose, FVector(), WorldToMetersScale);//
					// PlaneUpdate->LocalToTrackingTransform.ConcatenateRotation(FQuat(FVector::RightVector, -HALF_PI));
					PlaneUpdate->TrackingState = EARTrackingState::Tracking;
					UE_LOG(LogSpacesRuntime, Log, TEXT("Initializing %u plane TRACKING. Vertices: %d | Location: %s | Rotation: %s"), Plane.Key.D, PlaneUpdate->Vertices.Num(), *PlaneUpdate->LocalToTrackingTransform.GetLocation().ToString(), *PlaneUpdate->LocalToTrackingTransform.GetRotation().ToString());
				}
				else
				{
					PlaneUpdate->TrackingState = EARTrackingState::NotTracking;
					PlaneUpdate->LocalToTrackingTransform = FTransform::Identity;
					UE_LOG(LogSpacesRuntime, Log, TEXT("Initializing %u plane NOT TRACKING. Vertices: %d"), Plane.Key.D, PlaneUpdate->Vertices.Num());
				}
			}
		}
	}
    TrackedMeshHolder->EndMeshUpdates();
}

void FSpacesPlaneDetectionExtension::ClearPlanes() const
{
	RemovePlanes(PreviousPlanesId);
}

bool FSpacesPlaneDetectionExtension::CanEnableExtension() const
{
	return bDeviceSupportsPlaneDetection
	&& TrackedMeshHolder != nullptr
	;
}

bool FSpacesPlaneDetectionExtension::OnStart()
{
	ClearPlanes();
	return TryCreatePlaneDetection(CurrentSessionConfig->GetPlaneDetectionMode(), CurrentSessionConfig-> bUseConvexHullDetection);
}

void FSpacesPlaneDetectionExtension::OnPause()
{
}

void FSpacesPlaneDetectionExtension::OnResume()
{
}

bool FSpacesPlaneDetectionExtension::OnStop()
{
	PlaneDetection.Clear();
	return true;
}

FSpacesPlaneDetectionExtension::FSpacesPlaneDetectionExtension()
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Startup SpacesPlaneDetectionExtension"));
	RegisterOpenXRExtensionModularFeature();
}

FSpacesPlaneDetectionExtension::~FSpacesPlaneDetectionExtension()
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Shutdown SpacesPlaneDetectionExtension"));
	UnregisterOpenXRExtensionModularFeature();
}

void FSpacesPlaneDetectionExtension::PostCreateInstance(XrInstance InInstance)
{
	SetInstanceHandle(InInstance);
}

const void* FSpacesPlaneDetectionExtension::OnCreateSession(XrInstance InInstance, XrSystemId InSystem, const void* InNext)
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("OnCreateSession"));
	XrSystemPlaneDetectionPropertiesQCOM SystemPlaneDetectionProperties{
		XR_TYPE_SYSTEM_PLANE_DETECTION_PROPERTIES_QCOM
	};
	XrSystemProperties SystemProperties{XR_TYPE_SYSTEM_PROPERTIES,
	  &SystemPlaneDetectionProperties};
	if(XR_SUCCEEDED(OpenXRDynamicAPI::xrGetSystemProperties(InInstance, InSystem, &SystemProperties)))
	{
		//TODO(AS): Check for plane detection support when implemented
		// bDeviceSupportsPlaneDetection = SystemPlaneDetectionProperties.supportsPlaneDetection == XR_TRUE;
		bDeviceSupportsPlaneDetection = true;
	}
	
	UE_LOG(LogSpacesRuntime, Log, TEXT("bDeviceSupportsPlaneDetection: %s"), SnapdragonSpaces::BoolToStr(bDeviceSupportsPlaneDetection));

	XRTrackingSystem = GEngine->XRSystem.Get();
	
	if (IOpenXRARModule::IsAvailable())
	{
		TrackedMeshHolder = IOpenXRARModule::Get().GetTrackedMeshHolder();
		UE_LOG(LogSpacesRuntime, Log, TEXT("Obtained TrackedMeshHolder (%p)"), TrackedMeshHolder);
	}
	else
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("Cannot access to TrackedMeshHolder"));
	}
	
	return InNext;
}

void FSpacesPlaneDetectionExtension::PostCreateSession(XrSession InSession)
{
	SetSessionHandle(InSession);
}

FString FSpacesPlaneDetectionExtension::GetDisplayName()
{
	return FString(TEXT("Spaces Plane Detection"));
}

bool FSpacesPlaneDetectionExtension::GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions)
{
    OutExtensions.Add(XR_QCOM_PLANE_DETECTION_EXTENSION_NAME);
    return true;
}

void FSpacesPlaneDetectionExtension::UpdateDeviceLocations(XrSession InSession, XrTime DisplayTime, XrSpace TrackingSpace)
{
    if(!IsEnabled())
    {
	    return;
    }

	TMap<FGuid, FTrackedPlaneHolder> LocatedPlanes;
    if(!LocatePlanes(TrackingSpace, DisplayTime, LocatedPlanes)) return;

	TArray<FGuid> MissingPlanesId;
	CalculateMissingPlanes(LocatedPlanes, MissingPlanesId);
	
	RemovePlanes(MissingPlanesId);
	AllocatePlanes(LocatedPlanes);
	
}

bool FSpacesPlaneDetectionExtension::GetSupportedCaptureTypes(TArray<EARCaptureType>& OutTypes)
{
	OutTypes.Add(EARCaptureType::SpatialMapping);
	return true;
}

bool FSpacesPlaneDetectionExtension::OnEnabledCaptureType(EARCaptureType InType)
{
	return true;
}

bool FSpacesPlaneDetectionExtension::OnDisabledCaptureType(EARCaptureType InType)
{
	return true;
}

#if PLATFORM_ANDROID
void FSpacesPlaneDetectionExtension::GetRequiredPermissions(TArray<FString>& Permissions) const
{
	Permissions.Add("android.permission.CAMERA");
}
#endif

#undef LOCTEXT_NAMESPACE
